import os
import threading
try:
    import dill as pickle
except ImportError:
    import pickle

from os.path import join as opj

from .basejob import MultiJob
from .errors import PlamsError, FileError
from .functions import config, log

__all__ = ['JobManager']



class JobManager:
    """Class responsible for jobs and files management.

    Every instance has the following attributes:

    *   ``foldername`` -- the working folder name.
    *   ``workdir`` -- the absolute path to the working folder.
    *   ``logfile`` -- the absolute path to the logfile.
    *   ``input`` -- the absolute path to the copy of the input file in the working folder.
    *   ``settings`` -- a |Settings| instance for this job manager (see below).
    *   ``jobs`` -- a list of all jobs managed with this instance (in order of |run| calls).
    *   ``names`` -- a dictionary with names of jobs. For each name an integer value is stored indicating how many jobs with that basename have already been run.
    *   ``hashes`` -- a dictionary working as a hash-table for jobs.

    The *path* argument should be be a path to a directory inside which the main working folder will be created. If ``None``, the directory from where the whole script was executed is used.

    The ``foldername`` attribute is initially set to the *folder* argument. If such a folder already exists, the suffix ``.002`` is appended to *folder* and the number is increased (``.003``, ``.004``...) until a non-existsing name is found. If *folder* is ``None``, the name ``plams_workdir`` is used, followed by the same procedure to find a unique ``foldername``.

    The ``settings`` attribute is directly set to the value of *settings* argument (unlike in other classes where they are copied) and it should be a |Settings| instance with the following keys:

    *   ``hashing`` -- chosen hashing method (see |RPM|).
    *   ``counter_len`` -- length of number appended to the job name in case of a name conflict.
    *   ``remove_empty_directories`` -- if ``True``, all empty subdirectories of the working folder are removed on |finish|.

    """

    def __init__(self, settings, path=None, folder=None):

        self.settings = settings
        self.jobs = []
        self.names = {}
        self.hashes = {}

        if path is None:
            self.path = os.getcwd()
        elif os.path.isdir(path):
            self.path = os.path.abspath(path)
        else:
            raise PlamsError('Invalid path: {}'.format(path))

        basename = os.path.normpath(folder) if folder else 'plams_workdir'
        self.foldername = basename
        n = 2
        while os.path.exists(opj(self.path, self.foldername)):
            self.foldername = basename + '.' + str(n).zfill(3)
            n += 1

        self.workdir = opj(self.path, self.foldername)
        self.logfile = opj(self.workdir, 'logfile')
        self.input = opj(self.workdir, 'input')
        os.mkdir(self.workdir)



    def load_job(self, filename):
        """Load previously saved job from *filename*.

        *Filename* should be a path to a ``.dill`` file in some job folder. A |Job| instance stored there is loaded and returned. All attributes of this instance removed before pickling are restored. That includes ``jobmanager``, ``path`` (the absolute path to the folder containing *filename* is used) and ``default_settings`` (a list containing only ``config.job``).

        See |pickling| for details.
        """
        def setstate(job, path, parent=None):
            job.parent = parent
            job.jobmanager = self
            job.default_settings = [config.job]
            job.path = path
            if isinstance(job, MultiJob):
                job._lock = threading.Lock()
                for child in job:
                    setstate(child, opj(path, child.name), job)
                for otherjob in job.other_jobs():
                    setstate(otherjob, opj(path, otherjob.name), job)

            job.results.refresh()
            h = job.hash()
            if h is not None:
                self.hashes[h] = job
            for key in job._dont_pickle:
                job.__dict__[key] = None

        if os.path.isfile(filename):
            filename = os.path.abspath(filename)
        else:
            raise FileError('File {} not present'.format(filename))
        path = os.path.dirname(filename)
        with open(filename, 'rb') as f:
            try:
                job = pickle.load(f)
            except Exception as e:
                log("Unpickling of {} failed. Caught the following Exception:\n{}".format(filename, e), 1)
                return None

        setstate(job, path)
        return job



    def remove_job(self, job):
        """Remove *job* from the job manager. Forget its hash."""
        if job in self.jobs:
            self.jobs.remove(job)
            job.jobmanager = None
        h = job.hash()
        if h in self.hashes and self.hashes[h] == job:
            del self.hashes[h]
        if isinstance(job, MultiJob):
            for child in job:
                self.remove_job(child)
            for otherjob in job.other_jobs():
                self.remove_job(otherjob)



    def _register_name(self, job):
        """Register the name of the *job*.

        If a job with the same name was already registered, *job* is renamed by appending consecutive integers. The number of digits in the appended number is defined by the ``counter_len`` value in ``settings``.
        """

        name = job._full_name()
        if name in self.names:
            self.names[name] += 1
            job.name += '.'+ str(self.names[name]).zfill(self.settings.counter_len)
            log('Renaming job {} to {}'.format(name, job._full_name()), 3)
        else:
            self.names[name] = 1



    def _register(self, job):
        """Register the *job*. Register job's name (rename if needed) and create the job folder."""

        log('Registering job {}'.format(job.name), 7)
        job.jobmanager = self

        self._register_name(job)

        if job.path is None:
            if job.parent:
                job.path = opj(job.parent.path, job.name)
            else:
                job.path = opj(self.workdir, job.name)
        os.mkdir(job.path)

        self.jobs.append(job)
        job.status = 'registered'
        log('Job {} registered'.format(job.name), 7)



    def _check_hash(self, job):
        """Calculate the hash of *job* and, if it is not ``None``, search previously run jobs for the same hash. If such a job is found, return it. Otherwise, return ``None``"""
        h = job.hash()
        if h is not None:
            if h in self.hashes:
                prev = self.hashes[h]
                log('Job {} previously run as {}, using old results'.format(job.name, prev.name), 1)
                return prev
            else:
                self.hashes[h] = job
        return None



    def _clean(self):
        """Clean all registered jobs according to the ``save`` parameter in their ``settings``. If ``remove_empty_directories`` is ``True``,  traverse the working directory and delete all empty subdirectories."""
        log('Cleaning job manager', 7)

        for job in self.jobs:
            job.results._clean(job.settings.save)

        if self.settings.remove_empty_directories:
            for root, dirs, files in os.walk(self.workdir, topdown=False):
                for dirname in dirs:
                    fullname = opj(root, dirname)
                    if not os.listdir(fullname):
                        os.rmdir(fullname)

        log('Job manager cleaned', 7)





__all__ = ['add_Hs', 'apply_reaction_smarts', 'apply_template',
           'gen_coords_rdmol', 'get_backbone_atoms', 'modify_atom',
           'to_rdmol', 'from_rdmol', 'from_sequence', 'from_smiles', 'from_smarts',
           'partition_protein', 'readpdb', 'writepdb', 'get_substructure', 'get_conformations',
           'yield_coords']

"""
@author: Lars Ridder
@description: A set of functions to manipulate molecules based on RDKit

This is a series of functions that apply RDKit functionality on PLAMS molecules
"""

import sys
import random
from warnings import warn
try:
    import dill as pickle
except ImportError:
    import pickle

try:
    from rdkit import Chem, Geometry
    from rdkit.Chem import AllChem
except ImportError:
    __all__ = []

from ...mol.bond import Bond
from ...mol.atom import Atom
from ...mol.molecule import Molecule
from ...core.functions import add_to_class


def from_rdmol(rdkit_mol, confid=-1, properties=True):
    """
    Translate an RDKit molecule into a PLAMS molecule type.
    RDKit properties will be unpickled if their name ends with '_pickled'.

    :parameter rdkit_mol: RDKit molecule
    :type rdkit_mol: rdkit.Chem.Mol
    :parameter int confid: conformer identifier from which to take coordinates
    :parameter bool properties: If all Chem.Mol, Chem.Atom and Chem.Bond properties should be converted from RDKit to PLAMS format.
    :return: a PLAMS molecule
    :rtype: |Molecule|
    """
    if isinstance(rdkit_mol, Molecule):
        return rdkit_mol
    # Create PLAMS molecule
    plams_mol = Molecule()
    total_charge = 0
    try:
        Chem.Kekulize(rdkit_mol)
    except Exception:
        pass
    conf = rdkit_mol.GetConformer(id=confid)

    # Add atoms and assign properties to the PLAMS atom if *properties* = True
    for rd_atom in rdkit_mol.GetAtoms():
        pos = conf.GetAtomPosition(rd_atom.GetIdx())
        ch = rd_atom.GetFormalCharge()
        pl_atom = Atom(
            rd_atom.GetAtomicNum(), coords=(pos.x, pos.y, pos.z), charge=ch)
        if properties and rd_atom.GetPDBResidueInfo():
            pl_atom.properties.pdb_info = get_PDBResidueInfo(rd_atom)
        plams_mol.add_atom(pl_atom)
        total_charge += ch

        # Check for R/S information
        stereo = str(rd_atom.GetChiralTag())
        if stereo == 'CHI_TETRAHEDRAL_CCW':
            pl_atom.properties.stereo = 'counter-clockwise'
        elif stereo == 'CHI_TETRAHEDRAL_CW':
            pl_atom.properties.stereo = 'clockwise'

    # Add bonds to the PLAMS molecule
    for bond in rdkit_mol.GetBonds():
        at1 = plams_mol.atoms[bond.GetBeginAtomIdx()]
        at2 = plams_mol.atoms[bond.GetEndAtomIdx()]
        plams_mol.add_bond(Bond(at1, at2, bond.GetBondTypeAsDouble()))

        # Check for cis/trans information
        stereo, bond_dir = str(bond.GetStereo()), str(bond.GetBondDir())
        if stereo == 'STEREOZ' or stereo == 'STEREOCIS':
            plams_mol.bonds[-1].properties.stereo = 'Z'
        elif stereo == 'STEREOE' or stereo == 'STEREOTRANS':
            plams_mol.bonds[-1].properties.stereo = 'E'
        elif bond_dir == 'ENDUPRIGHT':
            plams_mol.bonds[-1].properties.stereo = 'up'
        elif bond_dir == 'ENDDOWNRIGHT':
            plams_mol.bonds[-1].properties.stereo = 'down'

    # Set charge and assign properties to PLAMS molecule and bonds if *properties* = True
    plams_mol.charge = total_charge
    if properties:
        prop_from_rdmol(plams_mol, rdkit_mol)
        for rd_atom, plams_atom in zip(rdkit_mol.GetAtoms(), plams_mol):
            prop_from_rdmol(plams_atom, rd_atom)
        for rd_bond, plams_bond in zip(rdkit_mol.GetBonds(), plams_mol.bonds):
            prop_from_rdmol(plams_bond, rd_bond)
    return plams_mol


def to_rdmol(plams_mol, sanitize=True, properties=True, assignChirality=False):
    """
    Translate a PLAMS molecule into an RDKit molecule type.
    PLAMS |Molecule|, |Atom| or |Bond| properties are pickled if they are neither booleans, floats,
    integers, floats nor strings, the resulting property names are appended with '_pickled'.

    :parameter plams_mol: A PLAMS molecule
    :parameter bool sanitize: Kekulize, check valencies, set aromaticity, conjugation and hybridization
    :parameter bool properties: If all |Molecule|, |Atom| and |Bond| properties should be converted from PLAMS to RDKit format.
    :parameter bool assignChirality: Assign R/S and cis/trans information, insofar as this was not yet present in the PLAMS molecule.
    :type plams_mol: |Molecule|
    :return: an RDKit molecule
    :rtype: rdkit.Chem.Mol
    """
    if isinstance(plams_mol, Chem.Mol):
        return plams_mol
    # Create rdkit molecule
    e = Chem.EditableMol(Chem.Mol())

    # Add atoms and assign properties to the RDKit atom if *properties* = True
    for pl_atom in plams_mol.atoms:
        rd_atom = Chem.Atom(int(pl_atom.atnum))
        if 'charge' in pl_atom.properties:
            rd_atom.SetFormalCharge(pl_atom.properties.charge)
        if properties:
            if 'pdb_info' in pl_atom.properties:
                set_PDBresidueInfo(rd_atom, pl_atom.properties.pdb_info)
            for prop in pl_atom.properties:
                if prop not in ('charge', 'pdb_info', 'stereo'):
                    prop_to_rdmol(pl_atom, rd_atom, prop)

        # Check for R/S information
        if pl_atom.properties.stereo:
            stereo = pl_atom.properties.stereo.lower()
            if stereo == 'counter-clockwise':
                rd_atom.SetChiralTag(Chem.rdchem.ChiralType.CHI_TETRAHEDRAL_CCW)
            elif stereo == 'clockwise':
                rd_atom.SetChiralTag(Chem.rdchem.ChiralType.CHI_TETRAHEDRAL_CW)
        e.AddAtom(rd_atom)

    # Mapping of PLAMS bond orders to RDKit bond types:
    def plams_to_rd_bonds(bo):
        if bo > 1.4 and bo < 1.6:
            return 12 # bond type for aromatic bond
        else:
            return int(bo)

    # Add bonds to the RDKit molecule
    for bond in plams_mol.bonds:
        a1 = plams_mol.atoms.index(bond.atom1)
        a2 = plams_mol.atoms.index(bond.atom2)
        e.AddBond(a1, a2, Chem.BondType(plams_to_rd_bonds(bond.order)))
    rdmol = e.GetMol()

    # Check for cis/trans information
    for pl_bond, rd_bond in zip(plams_mol.bonds, rdmol.GetBonds()):
        if pl_bond.properties.stereo:
            stereo = pl_bond.properties.stereo.lower()
            if stereo == 'e' or stereo == 'trans':
                rd_bond.SetStereo(Chem.rdchem.BondStereo.STEREOE)
            elif stereo == 'z' or stereo == 'cis':
                rd_bond.SetStereo(Chem.rdchem.BondStereo.STEREOZ)
            elif stereo == 'up':
                rd_bond.SetBondDir(Chem.rdchem.BondDir.ENDUPRIGHT)
            elif stereo == 'down':
                rd_bond.SetBondDir(Chem.rdchem.BondDir.ENDDOWNRIGHT)

    # Assign properties to RDKit molecule and bonds if *properties* = True
    if properties:
        for prop in plams_mol.properties:
            prop_to_rdmol(plams_mol, rdmol, prop)
        for pl_bond, rd_bond in zip(plams_mol.bonds, rdmol.GetBonds()):
            for prop in pl_bond.properties:
                if prop != 'stereo':
                    prop_to_rdmol(pl_bond, rd_bond, prop)

    if sanitize:
        Chem.SanitizeMol(rdmol)
    conf = Chem.Conformer()
    for i, atom in enumerate(plams_mol.atoms):
        xyz = Geometry.Point3D(atom._getx(), atom._gety(), atom._getz())
        conf.SetAtomPosition(i, xyz)
    rdmol.AddConformer(conf)
    # REB: Assign all stereochemistry, if it wasn't already there
    if assignChirality :
        Chem.rdmolops.AssignAtomChiralTagsFromStructure(rdmol,confId=conf.GetId(),replaceExistingTags=False)
        try :
            Chem.AssignStereochemistryFrom3D(rdmol,confId=conf.GetId(),replaceExistingTags=False)
        except AttributeError :
            pass
    return rdmol


pdb_residue_info_items = [
    'AltLoc', 'ChainId', 'InsertionCode', 'IsHeteroAtom', 'Name', 'Occupancy',
    'ResidueName', 'ResidueNumber', 'SecondaryStructure', 'SegmentNumber',
    'SerialNumber', 'TempFactor']
# 'MonomerType' was excluded because it is an rdkit type that cannot easilty be serialized


def get_PDBResidueInfo(rdkit_atom):
    pdb_info = {}
    for item in pdb_residue_info_items:
        get_function = 'Get' + item
        pdb_info[item] = rdkit_atom.GetPDBResidueInfo().__getattribute__(get_function)()
    return pdb_info


def set_PDBresidueInfo(rdkit_atom, pdb_info):
    atom_pdb_residue_info = Chem.AtomPDBResidueInfo()
    for item, value in pdb_info.items():
        set_function = 'Set' + item
        atom_pdb_residue_info.__getattribute__(set_function)(value)
    rdkit_atom.SetMonomerInfo(atom_pdb_residue_info)


def prop_to_rdmol(pl_obj, rd_obj, prop):
    """
    Convert a single PLAMS property into an RDKit property.

    :paramter pl_obj: A PLAMS object.
    :type pl_obj: |Molecule|, |Atom| or |Bond|.
    :parameter rd_obj: An RDKit object.
    :type rd_obj: rdkit.Chem.Mol, rdkit.Chem.Atom or rdkit.Chem.Bond
    :parameter str prop: The |Settings| key of the PLAMS property.
    """
    obj = type(pl_obj.properties.get(prop))
    obj_dict = {bool: rd_obj.SetBoolProp,
                float: rd_obj.SetDoubleProp,
                int: rd_obj.SetIntProp,
                str: rd_obj.SetProp}
    if obj_dict.get(obj):
        obj_dict[obj](prop, pl_obj.properties.get(prop))
    else:
        name = prop + '_pickled'
        try:
            rd_obj.SetProp(name, pickle.dumps(pl_obj.properties.get(prop), 0).decode())
        except (Exception, pickle.PicklingError):
            pass


def prop_from_rdmol(pl_obj, rd_obj):
    """
    Convert one or more RDKit properties into PLAMS properties.

    :paramter pl_obj: A PLAMS object.
    :type pl_obj: |Molecule|, |Atom| or |Bond|.
    :parameter rd_obj: An RDKit object.
    :type rd_obj: rdkit.Chem.Mol, rdkit.Chem.Atom or rdkit.Chem.Bond
    """
    prop_dict = rd_obj.GetPropsAsDict()
    for propname in prop_dict.keys():
        if propname == '__computedProps':
            continue
        if '_pickled' not in propname:
            pl_obj.properties[propname] = prop_dict[propname]
        else:
            prop = prop_dict[propname]
            propname = propname.rsplit('_pickled', 1)[0]
            pl_obj.properties[propname] = pickle.loads(prop.encode())


def from_smiles(smiles, nconfs=1, name=None, forcefield=None, rms=0.1):
    """
    Generates PLAMS molecule(s) from a smiles strings.

    :parameter str smiles: A smiles string
    :parameter int nconfs: Number of conformers to be generated
    :parameter str name: A name for the molecule
    :parameter str forcefield: Choose 'uff' or 'mmff' forcefield for geometry optimization
        and ranking of comformations. The default value None results in skipping of the
        geometry optimization step.
    :parameter float rms: Root Mean Square deviation threshold for
        removing similar/equivalent conformations
    :return: A molecule with hydrogens and 3D coordinates or a list of molecules if nconfs > 1
    :rtype: |Molecule| or list of PLAMS Molecules
    """
    smiles = str(smiles.split()[0])
    smiles = Chem.CanonSmiles(smiles)
    rdkit_mol = Chem.AddHs(Chem.MolFromSmiles(smiles))
    rdkit_mol.SetProp('smiles', smiles)
    return get_conformations(rdkit_mol, nconfs, name, forcefield, rms)


def from_smarts(smarts, nconfs=1, name=None, forcefield=None, rms=0.1):
    """
    Generates PLAMS molecule(s) from a smarts strings.
    This allows for example to define hydrogens explicitly.
    However it is less suitable for aromatic molecules (use from_smiles in that case).

    :parameter str smarts: A smarts string
    :parameter int nconfs: Number of conformers to be generated
    :parameter str name: A name for the molecule
    :parameter str forcefield: Choose 'uff' or 'mmff' forcefield for geometry
        optimization and ranking of comformations. The default value None results
        in skipping of the geometry optimization step.
    :parameter float rms: Root Mean Square deviation threshold for removing
        similar/equivalent conformations.
    :return: A molecule with hydrogens and 3D coordinates or a list of molecules if nconfs > 1
    :rtype: |Molecule| or list of PLAMS Molecules
    """
    smiles = str(smarts.split()[0])
    mol = Chem.MolFromSmarts(smiles)
    Chem.SanitizeMol(mol)
    molecule = Chem.AddHs(mol)
    molecule.SetProp('smiles', smiles)
    return get_conformations(molecule, nconfs, name, forcefield, rms)


def get_conformations(mol, nconfs=1, name=None, forcefield=None, rms=-1, enforceChirality=False):
    """
    Generates 3D conformation(s) for an rdkit_mol or a PLAMS Molecule

    :parameter mol: RDKit or PLAMS Molecule
    :type mol: rdkit.Chem.Mol or |Molecule|
    :parameter int nconfs: Number of conformers to be generated
    :parameter str name: A name for the molecule
    :parameter str forcefield: Choose 'uff' or 'mmff' forcefield for geometry
        optimization and ranking of comformations. The default value None results
        in skipping of the geometry optimization step
    :parameter float rms: Root Mean Square deviation threshold for removing
        similar/equivalent conformations.
    :parameter bool enforceChirality: Enforce the correct chirality if chiral centers are present
    :return: A molecule with hydrogens and 3D coordinates or a list of molecules if nconfs > 1
    :rtype: |Molecule| or list of PLAMS Molecules
    """

    if isinstance(mol, Molecule):
        rdkit_mol = to_rdmol(mol,assignChirality=enforceChirality)
    else:
        rdkit_mol = mol

    def MMFFenergy(cid):
        ff = AllChem.MMFFGetMoleculeForceField(
            rdkit_mol, AllChem.MMFFGetMoleculeProperties(rdkit_mol), confId=cid)
        try:
            energy = ff.CalcEnergy()
        except:
            msg = "MMFF energy calculation failed for molecule: " + Chem.MolToSmiles(rdkit_mol) + \
                  "\nNo geometry optimization was performed."
            warn(msg)
            energy = 1e9
        return energy

    def UFFenergy(cid):
        ff = AllChem.UFFGetMoleculeForceField(rdkit_mol, confId=cid)
        try:
            energy = ff.CalcEnergy()
        except:
            msg = "MMFF energy calculation failed for molecule: " + Chem.MolToSmiles(rdkit_mol) + \
                  "\nNo geometry optimization was performed."
            warn(msg)
            energy = 1e9
        return energy

    if name:
        rdkit_mol.SetProp('name', name)

    try:
        cids = list(AllChem.EmbedMultipleConfs(rdkit_mol,nconfs,pruneRmsThresh=rms,randomSeed=1,enforceChirality=enforceChirality))
    except:
         # ``useRandomCoords = True`` prevents (poorly documented) crash for large systems
        cids = list(AllChem.EmbedMultipleConfs(rdkit_mol,nconfs,pruneRmsThresh=rms,randomSeed=1,useRandomCoords=True,enforceChirality=enforceChirality))

    if forcefield:
        # Select the forcefield (UFF or MMFF)
        optimize_molecule, energy = {
            'uff': [AllChem.UFFOptimizeMolecule, UFFenergy],
            'mmff': [AllChem.MMFFOptimizeMolecule, MMFFenergy],
        }[forcefield]

        # Optimize and sort conformations
        for cid in cids:
            optimize_molecule(rdkit_mol, confId=cid)
        cids.sort(key=energy)

        # Remove duplicate conformations based on RMS
        if rms > 0:
            keep = [cids[0]]
            for cid in cids[1:]:
                for idx in keep:
                    try:
                        r = AllChem.AlignMol(rdkit_mol, rdkit_mol, cid, idx)
                    except:
                        r = rms + 1
                        message = "Alignment failed in multiple conformation generation: "
                        message += Chem.MolToSmiles(rdkit_mol)
                        message += "\nAssuming different conformations."
                        warn(message)
                    if r < rms:
                        break
                else:
                    keep.append(cid)
            cids = keep

    if nconfs == 1:
        return from_rdmol(rdkit_mol)
    else:
        return [from_rdmol(rdkit_mol, cid) for cid in cids]


def from_sequence(sequence, nconfs=1, name=None, forcefield=None, rms=0.1):
    """
    Generates PLAMS molecule from a peptide sequence.
    Includes explicit hydrogens and 3D coordinates.

    :parameter str sequence: A peptide sequence, e.g. 'HAG'
    :parameter int nconfs: Number of conformers to be generated
    :parameter str name: A name for the molecule
    :parameter str forcefield: Choose 'uff' or 'mmff' forcefield for geometry
        optimization and ranking of comformations. The default value None results
        in skipping of the geometry optimization step.
    :parameter float rms: Root Mean Square deviation threshold for removing
        similar/equivalent conformations.
    :return: A peptide molecule with hydrogens and 3D coordinates
        or a list of molecules if nconfs > 1
    :rtype: |Molecule| or list of PLAMS Molecules
    """
    rdkit_mol = Chem.AddHs(Chem.MolFromSequence(sequence))
    rdkit_mol.SetProp('sequence', sequence)
    return get_conformations(rdkit_mol, nconfs, name, forcefield, rms)


def calc_rmsd(mol1, mol2):
    """
    Superimpose two molecules and calculate the root-mean-squared deviations of
    the atomic positions.
    The molecules should be identical, but the ordering of the atoms may differ.

    :param mol1: Molecule 1
    :param mol2: Molecule 2
    :return: The rmsd after superposition
    :rtype: float
    """
    rdkit_mol1 = to_rdmol(mol1)
    rdkit_mol2 = to_rdmol(mol2)
    try:
        return AllChem.GetBestRMS(rdkit_mol1, rdkit_mol2)
    except:
        return -999


def modify_atom(mol, idx, element):
    """
    Change atom "idx" in molecule "mol" to "element" and add or remove hydrogens accordingly

    :parameter mol: molecule to be modified
    :type mol: |Molecule| or rdkit.Chem.Mol
    :parameter int idx: index of the atom to be modified
    :parameter str element:
    :return: Molecule with new element and possibly added or removed hydrogens
    :rtype: |Molecule|
    """
    rdmol = to_rdmol(mol)
    if rdmol.GetAtomWithIdx(idx).GetSymbol() == element:
        return mol
    else:
        e = Chem.EditableMol(rdmol)
        for neighbor in reversed(rdmol.GetAtomWithIdx(idx - 1).GetNeighbors()):
            if neighbor.GetSymbol() == 'H':
                e.RemoveAtom(neighbor.GetIdx())
        e.ReplaceAtom(idx - 1, Chem.Atom(element))
        newmol = e.GetMol()
        Chem.SanitizeMol(newmol)
        newmol = Chem.AddHs(newmol, addCoords=True)
        return from_rdmol(newmol)


def apply_template(mol, template):
    """
    Modifies bond orders in PLAMS molecule according template smiles structure.

    :parameter mol: molecule to be modified
    :type mol: |Molecule| or rdkit.Chem.Mol
    :parameter str template: smiles string defining the correct chemical structure
    :return: Molecule with correct chemical structure and provided 3D coordinates
    :rtype: |Molecule|
    """
    rdmol = to_rdmol(mol, sanitize=False)
    template_mol = Chem.AddHs(Chem.MolFromSmiles(template))
    newmol = Chem.AllChem.AssignBondOrdersFromTemplate(template_mol, rdmol)
    return from_rdmol(newmol)


def apply_reaction_smarts(
        mol, reaction_smarts, complete=False, forcefield=None, return_rdmol=False):
    """
    Applies reaction smirks and returns product.
    If returned as a PLAMS molecule, thismolecule.properties.orig_atoms
    is a list of indices of atoms that have not been changed
    (which can for example be used partially optimize new atoms only with the freeze keyword)

    :parameter mol: molecule to be modified
    :type mol: |Molecule| or rdkit.Chem.Mol
    :parameter str reactions_smarts: Reactions smarts to be applied to molecule
    :parameter complete: Apply reaction until no further changes occur or given
        fraction of reaction centers have been modified
    :type complete: bool or float (value between 0 and 1)
    :parameter forcefield: Specify 'uff' or 'mmff' to apply forcefield based
        geometry optimization of product structures.
    :type forcefield: str
    :param bool return_rdmol: return a RDKit molecule if true, otherwise a PLAMS molecule
    :return: (product molecule, list of unchanged atoms)
    :rtype: (|Molecule|, list of int)
    """
    def react(reactant, reaction):
        """ Apply reaction to reactant and return products """
        ps = reaction.RunReactants([reactant])
        # if reaction doesn't apply, return the reactant
        if len(ps) == 0:
            return [(reactant, range(reactant.GetNumAtoms()))]
        full = len(ps)
        while complete:  # when complete is True
            # apply reaction until no further changes
            r = random.randint(0, len(ps) - 1)
            reactant = ps[r][0]
            ps = reaction.RunReactants([reactant])
            if len(ps) == 0 or len(ps) / full < (1 - complete):
                ps = [[reactant]]
                break
        # add hydrogens and generate coordinates for new atoms
        products = []
        for p in ps[0]:
            Chem.SanitizeMol(p)
            q = Chem.AddHs(p)
            Chem.SanitizeMol(q)
            u = gen_coords_rdmol(q)  # These are the atoms that have not changed
            products.append((q, u))
        return products

    mol = to_rdmol(mol)
    reaction = AllChem.ReactionFromSmarts(reaction_smarts)
    # RDKit removes fragments that are disconnected from the reaction center
    # In order to keep these, the molecule is first split in separate fragments
    # and the results, including non-reacting parts, are re-combined afterwards
    frags = (Chem.GetMolFrags(mol, asMols=True))
    product = Chem.Mol()
    unchanged = []  # List of atoms that have not changed
    for frag in frags:
        for p, u in react(frag, reaction):
            unchanged += [product.GetNumAtoms() + i for i in u]
            product = Chem.CombineMols(product, p)
    if forcefield:
        optimize_coordinates(product, forcefield, fixed=unchanged)
    # The molecule is returned together with a list of atom indices of the atoms
    # that are identical to those
    # in the reactants. This list can be used in subsequent partial optimization of the molecule
    if not return_rdmol:
        product = from_rdmol(product)
        product.properties.orig_atoms = [a + 1 for a in unchanged]
    return product


def gen_coords(plamsmol):
    """ Calculate 3D positions only for atoms without coordinates """
    rdmol = to_rdmol(plamsmol)
    unchanged = gen_coords_rdmol(rdmol)
    conf = rdmol.GetConformer()
    for a in range(len(plamsmol.atoms)):
        pos = conf.GetAtomPosition(a)
        atom = plamsmol.atoms[a]
        atom._setx(pos.x)
        atom._sety(pos.y)
        atom._setz(pos.z)
    return [a + 1 for a in unchanged]


def gen_coords_rdmol(rdmol):
    ref = rdmol.__copy__()
    conf = rdmol.GetConformer()
    coordDict = {}
    unchanged = []
    maps = []
    # Put known coordinates in coordDict
    for i in range(rdmol.GetNumAtoms()):
        pos = conf.GetAtomPosition(i)
        if (-0.0001 < pos.x < 0.0001) and (-0.0001 < pos.y < 0.0001) and \
           (-0.0001 < pos.z < 0.0001):
            continue  # atom without coordinates
        coordDict[i] = pos
        unchanged.append(i)
        maps.append((i, i))
    # compute coordinates for new atoms, keeping known coordinates
    rms = 1
    rs = 1
    # repeat embedding and alignment until the rms of mapped atoms is sufficiently small
    if rdmol.GetNumAtoms() > len(maps):
        while rms > 0.1:
            AllChem.EmbedMolecule(rdmol, coordMap=coordDict, randomSeed=rs,
                                  useBasicKnowledge=True)
            # align new molecule to original coordinates
            rms = AllChem.AlignMol(rdmol, ref, atomMap=maps)
            rs += 1
    return unchanged


def optimize_coordinates(rdkit_mol, forcefield, fixed=[]):
    def MMFFminimize():
        ff = AllChem.MMFFGetMoleculeForceField(
            rdkit_mol, AllChem.MMFFGetMoleculeProperties(rdkit_mol))
        for f in fixed:
            ff.AddFixedPoint(f)
        try:
            ff.Minimize()
        except:
            warn("MMFF geometry optimization failed for molecule: " + Chem.MolToSmiles(rdkit_mol))

    def UFFminimize():
        ff = AllChem.UFFGetMoleculeForceField(rdkit_mol, ignoreInterfragInteractions=True)
        for f in fixed:
            ff.AddFixedPoint(f)
        try:
            ff.Minimize()
        except:
            warn("UFF geometry optimization failed for molecule: " + Chem.MolToSmiles(rdkit_mol))
    optimize_molecule = {
        'uff': UFFminimize,
        'mmff': MMFFminimize}[forcefield]
    Chem.SanitizeMol(rdkit_mol)
    optimize_molecule()
    return


def write_molblock(plams_mol, file=sys.stdout):
    file.write(Chem.MolToMolBlock(to_rdmol(plams_mol)))


def readpdb(pdb_file, removeHs=False, proximityBonding=False, return_rdmol=False):
    """
    Generate a molecule from a PDB file

    :param pdb_file: The PDB file to read
    :type pdb_file: path- or file-like
    :param bool removeHs: Hydrogens are removed if True
    :param bool proximityBonding: Enables automatic proximity bonding
    :param bool return_rdmol: return a RDKit molecule if true, otherwise a PLAMS molecule
    :return: The molecule
    :rtype: |Molecule| or rdkit.Chem.Mol
    """
    try:
        pdb_file = open(pdb_file, 'r')
    except TypeError:
        pass  # pdb_file is a file-like object... hopefully

    pdb_mol = Chem.MolFromPDBBlock(pdb_file.read(), removeHs=removeHs, proximityBonding=proximityBonding)
    return pdb_mol if return_rdmol else from_rdmol(pdb_mol)


def writepdb(mol, pdb_file=sys.stdout):
    """
    Write a PDB file from a molecule

    :parameter mol: molecule to be exported to PDB
    :type mol: |Molecule| or rdkit.Chem.Mol
    :param pdb_file: The PDB file to write to, or a filename
    :type pdb_file: path- or file-like
    """
    try:
        pdb_file = open(pdb_file, 'w')
    except TypeError:
        pass  # pdb_file is a file-like object... hopefully

    mol = to_rdmol(mol)
    pdb_file.write(Chem.MolToPDBBlock(mol))


def add_Hs(mol, forcefield=None, return_rdmol=False):
    """
    Add hydrogens to protein molecules read from PDB.
    Makes sure that the hydrogens get the correct PDBResidue info.

    :param mol: Molecule to be protonated
    :type mol: |Molecule| or rdkit.Chem.Mol
    :param str forcefield: Specify 'uff' or 'mmff' to apply forcefield based
        geometry optimization on new atoms.
    :param bool return_rdmol: return a RDKit molecule if true, otherwise a PLAMS molecule
    :return: A molecule with explicit hydrogens added
    :rtype: |Molecule| or rdkit.Chem.Mol
    """
    mol = to_rdmol(mol)
    retmol = Chem.AddHs(mol)
    for atom in retmol.GetAtoms():
        if atom.GetPDBResidueInfo() is None and atom.GetSymbol() == 'H':
            bond = atom.GetBonds()[0]
            if bond.GetBeginAtom().GetIdx() == atom.GetIdx:
                connected_atom = bond.GetEndAtom()
            else:
                connected_atom = bond.GetBeginAtom()
            try:
                ResInfo = connected_atom.GetPDBResidueInfo()
                if ResInfo is None:
                    continue  # Segmentation faults are raised if ResInfo is None
                atom.SetMonomerInfo(ResInfo)
                atomname = 'H' + atom.GetPDBResidueInfo().GetName()[1:]
                atom.GetPDBResidueInfo().SetName(atomname)
            except:
                pass
    unchanged = gen_coords_rdmol(retmol)
    if forcefield:
        optimize_coordinates(retmol, forcefield, fixed=unchanged)
    return retmol if return_rdmol else from_rdmol(retmol)


def add_fragment(rwmol, frag, rwmol_atom_idx=None, frag_atom_idx=None,
                 bond_order=None):
    molconf = rwmol.GetConformer()
    fragconf = frag.GetConformer()
    new_indices = []
    for a in frag.GetAtoms():
        new_index = rwmol.AddAtom(a)
        new_indices.append(new_index)
        molconf.SetAtomPosition(new_index, fragconf.GetAtomPosition(a.GetIdx()))
    for b in frag.GetBonds():
        ba = b.GetBeginAtomIdx()
        ea = b.GetEndAtomIdx()
        rwmol.AddBond(new_indices[ba], new_indices[ea], b.GetBondType())
    if bond_order:
        rwmol.AddBond(rwmol_atom_idx, new_indices[frag_atom_idx],
                      Chem.BondType.values[bond_order])
        rwmol.GetAtomWithIdx(new_indices[frag_atom_idx]).SetNumRadicalElectrons(0)


def get_fragment(mol, indices, incl_expl_Hs=True, neutralize=True):
    molconf = mol.GetConformer()
    fragment = Chem.RWMol(Chem.Mol())
    fragconf = Chem.Conformer()
    # Put atoms in fragment
    for i in indices:
        atom = mol.GetAtomWithIdx(i)
        new_index = fragment.AddAtom(atom)
        pos = molconf.GetAtomPosition(i)
        fragconf.SetAtomPosition(new_index, pos)
    # Put bonds in fragment
    for b in mol.GetBonds():
        ba = b.GetBeginAtomIdx()
        ea = b.GetEndAtomIdx()
        if ba in indices and ea in indices:
            fragment.AddBond(indices.index(ba), indices.index(ea),
                             b.GetBondType())
            continue
        if not incl_expl_Hs:
            continue
        if ba in indices and mol.GetAtomWithIdx(ea).GetSymbol() == 'H':
            hi = fragment.AddAtom(mol.GetAtomWithIdx(ea))
            fragconf.SetAtomPosition(hi, molconf.GetAtomPosition(ea))
            fragment.AddBond(indices.index(ba), hi, Chem.BondType.SINGLE)
            continue
        if ea in indices and mol.GetAtomWithIdx(ba).GetSymbol() == 'H':
            hi = fragment.AddAtom(mol.GetAtomWithIdx(ba))
            fragconf.SetAtomPosition(hi, molconf.GetAtomPosition(ba))
            fragment.AddBond(indices.index(ea), hi, Chem.BondType.SINGLE)
    ret_frag = fragment.GetMol()
    Chem.SanitizeMol(ret_frag)
    if neutralize:
        for atom in ret_frag.GetAtoms():
            nrad = atom.GetNumRadicalElectrons()
            if nrad > 0:
                atom.SetNumExplicitHs(atom.GetNumExplicitHs() + nrad)
                atom.SetNumRadicalElectrons(0)
    Chem.SanitizeMol(ret_frag)
    ret_frag.AddConformer(fragconf)
    return ret_frag


def partition_protein(mol, residue_bonds=None, split_heteroatoms=True, return_rdmol=False):
    """
    Splits a protein molecule into capped amino acid fragments and caps.

    :param mol: A protein molecule
    :type mol: |Molecule| or rdkit.Chem.Mol
    :param tuple residue_bonds: a tuple of pairs of residue number indicating which
        peptide bonds to split. If none, split all peptide bonds.
    :param bool split_heteroatoms: if True, all bonds between a heteroatom and
        a non-heteroatom across residues are removed
    :return: list of fragments, list of caps
    """
    mol = to_rdmol(mol)
    caps = []
    em = Chem.RWMol(mol)
    if split_heteroatoms:
        for bond in mol.GetBonds():
            resinfa = bond.GetBeginAtom().GetPDBResidueInfo()
            resinfb = bond.GetEndAtom().GetPDBResidueInfo()
            if resinfa.GetIsHeteroAtom() is not resinfb.GetIsHeteroAtom():
                if resinfa.GetResidueNumber() != resinfb.GetResidueNumber():
                    em.RemoveBond(bond.GetBeginAtomIdx(), bond.GetEndAtomIdx())
    # Split peptide bonds
    pept_bond = Chem.MolFromSmarts('[C;X4;H1,H2][CX3](=O)[NX3][C;X4;H1,H2][CX3](=O)')
    for match in mol.GetSubstructMatches(pept_bond):
        if residue_bonds:
            resa = mol.GetAtomWithIdx(match[1]).GetPDBResidueInfo().GetResidueNumber()
            resb = mol.GetAtomWithIdx(match[3]).GetPDBResidueInfo().GetResidueNumber()
            if (resa, resb) not in residue_bonds and (resb, resa) not in residue_bonds:
                continue
        cap = get_fragment(mol, match[0:5])
        cap = add_Hs(cap, return_rdmol=True)
        caps.append(cap if return_rdmol else from_rdmol(cap))
        cap_o_ind = cap.GetSubstructMatch(Chem.MolFromSmarts('[C;X4][CX3]=O'))
        cap_o = get_fragment(cap, cap_o_ind, neutralize=False)
        cap_n_ind = cap.GetSubstructMatch(Chem.MolFromSmarts('O=[CX3][NX3][C;X4]'))[2:]
        cap_n = get_fragment(cap, cap_n_ind, neutralize=False)
        em.RemoveBond(match[1], match[3])
        add_fragment(em, cap_o, match[3], 1, 1)
        add_fragment(em, cap_n, match[1], 0, 1)
    # Split disulfide bonds
    ss_bond = Chem.MolFromSmarts('[C;X4;H1,H2]SS[C;X4;H1,H2]')
    for match in mol.GetSubstructMatches(ss_bond):
        cap = get_fragment(mol, match[0:5])
        cap = add_Hs(cap, return_rdmol=True)
        caps.append(cap if return_rdmol else from_rdmol(cap))
        cap_s_ind = cap.GetSubstructMatch(Chem.MolFromSmarts('[C;X4]SS[C;X4]'))
        cap_s1 = get_fragment(cap, cap_s_ind[0:2], neutralize=False)
        cap_s2 = get_fragment(cap, cap_s_ind[2:4], neutralize=False)
        em.RemoveBond(match[1], match[2])
        add_fragment(em, cap_s1, match[2], 1, 1)
        add_fragment(em, cap_s2, match[1], 0, 1)
    frags = Chem.GetMolFrags(em.GetMol(), asMols=True, sanitizeFrags=False)
    if not return_rdmol:
        frags = [from_rdmol(frag) for frag in frags]
    return frags, caps


def charge_AAs(mol, return_rdmol=False):
    ionizations = {
        'ARG_NH2': 1,
        'LYS_NZ': 1,
        'GLU_OE2': -1,
        'ASP_OD2': -1}
    mol = to_rdmol(mol)
    for atom in mol.GetAtoms():
        resinfo = atom.GetPDBResidueInfo()
        res_atom = resinfo.GetResidueName() + '_' + resinfo.GetName().strip()
        try:
            atom.SetFormalCharge(ionizations[res_atom])
            Chem.SanitizeMol(mol)
        except KeyError:
            pass
        Chem.SanitizeMol(mol)
    return mol if return_rdmol else from_rdmol(mol)


def get_backbone_atoms(mol):
    """
    Return a list of atom indices corresponding to the backbone atoms in a peptide molecule.
    This function assumes PDB information in properties.pdb_info of each atom, which is the case
    if the molecule is generated with the "readpdb" or "from_sequence" functions.

    :parameter mol: a peptide molecule
    :type mol: |Molecule| or rdkit.Chem.Mol
    :return: a list of atom indices
    :rtype: list
    """
    mol = from_rdmol(mol)
    backbone = ['N', 'CA', 'C', 'O']
    return [a for a in range(1, len(mol) + 1)
            if str(mol[a].properties.pdb_info.Name).strip() in backbone]


def get_substructure(mol, func_list):
    """
    Search for functional groups within a molecule based on a list of reference functional groups.
    SMILES strings, PLAMS and/or RDKit molecules can be used interchangeably in "func_list".

    Example:

    .. code:: python

        >>> mol = from_smiles('OCCO')  # Ethylene glycol
        >>> func_list = ['[H]O', 'C[N+]', 'O=PO']
        >>> get_substructure(mol, func_list)

        {'[H]O': [(<scm.plams.mol.atom.Atom at 0x125183518>,
                   <scm.plams.mol.atom.Atom at 0x1251836a0>),
                  (<scm.plams.mol.atom.Atom at 0x125183550>,
                   <scm.plams.mol.atom.Atom at 0x125183240>)]}

    :parameter mol: A PLAMS molecule.
    :type mol: |Molecule|
    :parameter list func_list: A list of functional groups.
        Functional groups can be represented by SMILES strings, PLAMS and/or RDKit molecules.
    :return: A dictionary with functional groups from "func_list" as keys and a list of n-tuples
        with matching PLAMS |Atom| as values.
    """
    def _to_rdmol(functional_group):
        """ Turn a SMILES strings, RDKit or PLAMS molecules into an RDKit molecule. """
        if isinstance(functional_group, str):
            # RDKit tends to remove explicit hydrogens if SANITIZE_ADJUSTHS is enabled
            sanitize = Chem.SanitizeFlags.SANITIZE_ALL^Chem.SanitizeFlags.SANITIZE_ADJUSTHS
            ret = Chem.MolFromSmiles(functional_group, sanitize=False)
            Chem.rdmolops.SanitizeMol(ret, sanitizeOps=sanitize)
            return ret
        elif isinstance(functional_group, Molecule):
            return to_rdmol(functional_group)
        elif isinstance(functional_group, Chem.Mol):
            return functional_group
        raise TypeError('get_substructure: ' + str(type(functional_group)) + ' is not a supported \
                        object type')

    def _get_match(mol, rdmol, functional_group):
        """ Perform a substructure match on "mol".
        If a match is found, return a list of n-tuples consisting PLAMS |Atom|.
        Otherwise return False. """
        matches = rdmol.GetSubstructMatches(functional_group)
        if matches:
            return [tuple(mol[j+1] for j in idx_tup) for idx_tup in matches]
        return False

    rdmol = to_rdmol(mol)
    rdmol_func_list = [_to_rdmol(i) for i in func_list]
    gen = (_get_match(mol, rdmol, i) for i in rdmol_func_list)
    return {key: value for key, value in zip(func_list, gen) if value}

  
def yield_coords(rdmol, id=-1):
    """Take an rdkit molecule and yield its coordinates as 3-tuples.

    .. code-block:: python

        >>> from scm.plams import yield_coords
        >>> from rdkit import Chem

        >>> rdmol = Chem.Mol(...)  # e.g. Methane
        >>> for xyz in yield_coords(rdmol):
        ...     print(xyz)
        (-0.0, -0.0, -0.0)
        (0.6405, 0.6405, -0.6405)
        (0.6405, -0.6405, 0.6405)
        (-0.6405, 0.6405, 0.6405)
        (-0.6405, -0.6405, -0.6405)


    The iterator produced by this function can, for example, be passed to
    :meth:`Molecule.from_array()<scm.plams.mol.molecule.Molecule.from_array>`
    the update the coordinates of a PLAMS Molecule in-place.

    .. code-block:: python

        >>> from scm.plams import Molecule

        >>> mol = Molecule(...)

        >>> xyz_iterator = yield_coords(rdmol)
        >>> mol.from_array(xyz_iterator)


    :parameter rdmol: An RDKit mol.
    :type rdmol: rdkit.Chem.Mol
    :parameter int id: The ID of the desired conformer.
    :return: An iterator yielding 3-tuples with *rdmol*'s Cartesian coordinates.
    :rtype: iterator
    """
    conf = rdmol.GetConformer(id=id)
    for atom in rdmol.GetAtoms():
        pos = conf.GetAtomPosition(atom.GetIdx())
        yield (pos.x, pos.y, pos.z)

        
@add_to_class(Molecule)
def assign_chirality(self):
    """
    Assigns stereo-info to PLAMS molecule by invoking RDKIT
    """
    rd_mol = to_rdmol(self, assignChirality=True)
    pl_mol = from_rdmol(rd_mol)

    # Add R/S info to self
    for iat,pl_atom in enumerate(pl_mol.atoms):
        # Check for R/S information
        if pl_atom.properties.stereo:
            self.atoms[iat].properties.stereo = pl_atom.properties.stereo

    # Add cis/trans information to self
    for ibond,pl_bond in enumerate(pl_mol.bonds):
        if pl_bond.properties.stereo:
            self.bonds[ibond] = pl_bond.properties.stereo

            
@add_to_class(Molecule)
def get_chirality(self):
    """
    Returns the chirality of the atoms
    """
    rd_mol = to_rdmol(self, assignChirality=True)
    return Chem.FindMolChiralCenters(rd_mol,force=True,includeUnassigned=True)



from collections import OrderedDict
from ..core.functions import add_to_instance
from ..core.basejob import MultiJob
from ..core.results import Results
from ..core.settings import Settings
from ..mol.molecule import Molecule
from ..mol.atom import Atom
from ..interfaces.adfsuite.ams import AMSJob
from ..tools.units import Units

__all__ = ['CounterpoiseEnergyJob', 'CounterpoiseEnergyResults']

class CounterpoiseEnergyResults(Results):
    """Results class for counterpoise corrected interaction and binding energies.
    """
    def get_all_energies(self, unit='au'):
        """'A_geo_AB_basis_AB' should be read as "atoms A at the AB geometry with the AB basis set" 
            BSSE is the basis set superposition error
            Eint is the interaction energy
            Ebind is the binding energy (the complex relative to relaxed and isolated A and B)
            _raw means no counterpoise correction
            _cp means counterpoise-corrected
            Edef is the deformation energy
        """
        E = {}
        
        E['AB_geo_AB_basis_AB'] = self.job.AB.results.get_energy(unit=unit)
        E['A_geo_A_basis_A'] = Units.convert(self.job.A, 'hartree', unit) if isinstance(self.job.A,float) else self.job.children['A_geo_A_basis_A'].results.get_energy(unit=unit)
        E['B_geo_B_basis_B'] = Units.convert(self.job.B, 'hartree', unit) if isinstance(self.job.B,float) else self.job.children['B_geo_B_basis_B'].results.get_energy(unit=unit)
            
        E['A_geo_AB_basis_AB'] = self.job.children['A_geo_AB_basis_AB'].results.get_energy(unit=unit)
        E['A_geo_AB_basis_A'] = self.job.children['A_geo_AB_basis_A'].results.get_energy(unit=unit)
        E['B_geo_AB_basis_AB'] = self.job.children['B_geo_AB_basis_AB'].results.get_energy(unit=unit)
        E['B_geo_AB_basis_B'] = self.job.children['B_geo_AB_basis_B'].results.get_energy(unit=unit)


        E['BSSE_A'] = E['A_geo_AB_basis_AB'] - E['A_geo_AB_basis_A']
        E['BSSE_B'] = E['B_geo_AB_basis_AB'] - E['B_geo_AB_basis_B']

        E['BSSE_tot'] = E['BSSE_A'] + E['BSSE_B']

        E['Eint_raw'] = E['AB_geo_AB_basis_AB'] - E['A_geo_AB_basis_A'] - E['B_geo_AB_basis_B']
        E['Eint_cp'] = E['AB_geo_AB_basis_AB'] - E['A_geo_AB_basis_AB'] - E['B_geo_AB_basis_AB']

        E['Edef_A'] = E['A_geo_AB_basis_A'] - E['A_geo_A_basis_A']
        E['Edef_B'] = E['B_geo_AB_basis_B'] - E['B_geo_B_basis_B']

        E['Ebind_raw'] = E['AB_geo_AB_basis_AB'] - E['A_geo_A_basis_A'] - E['B_geo_B_basis_B']
        E['Ebind_cp'] = E['Eint_cp'] + E['Edef_A'] + E['Edef_B']

        return E


class CounterpoiseEnergyJob(MultiJob):
    """A class for calculating the counterpoise-corrected interaction and binding energies.
    """

    _result_type = CounterpoiseEnergyResults

    def __init__(self, AB, ids_state_A, settings_state_AB=None, A=None, B=None, settings_state_A=None, settings_state_B=None, **kwargs):
        """
        AB is either a Molecule or a previously run AMSJob (from which the molecule will be extracted with .results.get_main_molecule())
        If it is a Molecule, a job will be created with settings_state_AB (default task Geometry Optimization).
        If it is an AMSJob, the structure and energy will be read from that job. settings_state_AB will be ignored.

        ids_state_A is a list of one-based atom indices making up the first molecule A. 
        All other atoms are assigned to belong to the second molecule B.

        Set A and B to the energies (in Hartree) of the isolated fragments.
        Alternatively, set A and B to AMSJobs instances or Molecules of previously run geometry optimizations for the isolated molecules, which will then be taken as starting points for the geometry optimization. If they are None, the structure from the complex AB will be taken as a starting point for the geometry optimization.

        settings_state_A: Setting object for molecule A. Defaults to the same as settings_state_AB. The Task will be applied only to the isolated molecule A, if such a job is run.
        settings_state_B: Setting object for molecule B. Defaults to the same as settings_state_AB. The Task will be applied only to the isolated molecule B, if such job is run.

        The calculations involving ghost atoms are enforced to be single point calculations.

        kwargs: other options to be passed to the MultiJob constructor (for example the name)
        """
        MultiJob.__init__(self, children=OrderedDict(), **kwargs)

        self.A = A
        self.B = B
        self.AB = AB

        # copy the settings so that we wont modify the ones provided as input by the user
        settings_state_AB = settings_state_AB.copy() if settings_state_AB else Settings()
        if settings_state_A:
            settings_state_A = settings_state_A.copy()
        else:
            settings_state_A = settings_state_AB.copy()
            settings_state_A.input.ams.Task = 'GeometryOptimization'
        if settings_state_B:
            settings_state_B = settings_state_B.copy()
        else:
            settings_state_B = settings_state_AB.copy()
            settings_state_B.input.ams.Task = 'GeometryOptimization'

        # In case the charge key is not specified, explicitly set the value to 0.
        # This is to prevent the charge in molecule.properties.charge (set by get_main_molecule())
        # to be used in case of neutral systems
        for s in [settings_state_AB, settings_state_A, settings_state_B]:
            if not 'charge' in s.input.ams.system:
                s.input.ams.system.charge = 0

        if isinstance(self.AB, Molecule):
            self.children['AB_geo_AB_basis_AB'] = AMSJob(molecule=self.AB, settings=settings_state_AB, name='AB_geo_AB_basis_AB')
            main_child = self.children['AB_geo_AB_basis_AB']
            self.AB = main_child
        elif isinstance(AB, AMSJob):
            main_child = self.AB
        else:
            raise ValueError('Unknown type for argument AB: {} (must be Molecule or AMSJob)'.format(type(self.AB)))

        name = 'A_geo_AB_basis_AB'
        s = settings_state_A.copy()
        s.input.ams.Task = 'SinglePoint'
        job = AMSJob(settings=s, name=name)
        @add_to_instance(job)
        def prerun(self):
            mol = main_child.results.get_main_molecule()
            for i in range(1, len(mol)+1):
                mol[i].properties.ghost = not i in ids_state_A
            self.molecule = mol
        self.children[name] = job

        name = 'A_geo_AB_basis_A'
        s = settings_state_A.copy()
        s.input.ams.Task = 'SinglePoint'
        job = AMSJob(settings=s, name=name)
        @add_to_instance(job)
        def prerun(self):
            mainmol = main_child.results.get_main_molecule()
            mol = Molecule()
            mol.lattice = mainmol.lattice
            for i in range(1, len(mainmol)+1):
                if i in ids_state_A:
                    mol.add_atom(Atom(atnum=mainmol[i].atnum, coords=mainmol[i].coords))
            self.molecule = mol
        self.children[name] = job


        name = 'B_geo_AB_basis_AB'
        s = settings_state_B.copy()
        s.input.ams.Task = 'SinglePoint'
        job = AMSJob(settings=s, name=name)
        @add_to_instance(job)
        def prerun(self):
            mol = main_child.results.get_main_molecule()
            for i in range(1, len(mol)+1):
                mol[i].properties.ghost = i in ids_state_A
            self.molecule = mol
        self.children[name] = job

        name = 'B_geo_AB_basis_B'
        s = settings_state_B.copy()
        s.input.ams.Task = 'SinglePoint'
        job = AMSJob(settings=s, name=name)
        @add_to_instance(job)
        def prerun(self):
            mainmol = main_child.results.get_main_molecule()
            mol = Molecule()
            mol.lattice = mainmol.lattice
            for i in range(1, len(mainmol)+1):
                if i not in ids_state_A:
                    mol.add_atom(Atom(atnum=mainmol[i].atnum, coords=mainmol[i].coords))
            self.molecule = mol
        self.children[name] = job

        def add_isolated_child(X, name, settings):
            s = settings.copy()
            job = AMSJob(settings=s, name=name)
            if isinstance(X, Molecule):
                job.molecule = X
            else:
                @add_to_instance(job)
                def prerun(self):
                    mainmol = main_child.results.get_main_molecule()
                    mol = Molecule()
                    mol.lattice = mainmol.lattice
                    for i in range(1, len(mainmol)+1):
                        if (name.startswith('A') and i in ids_state_A) or (name.startswith('B') and i not in ids_state_A):
                            mol.add_atom(Atom(atnum=mainmol[i].atnum, coords=mainmol[i].coords))
                    self.molecule = mol
            self.children[name] = job

        if not isinstance(self.A, float):
            add_isolated_child(self.A, 'A_geo_A_basis_A', settings_state_A)
        if not isinstance(self.B, float):
            add_isolated_child(self.B, 'B_geo_B_basis_B', settings_state_B)
        




# Perform a geometry optimization of a water molecule and compute
# the vibrational normal modes using GFN1-xTB.

# You could also load the geometry from an xyz file: 
# molecule = Molecule('path/my_molecule.xyz')
molecule = Molecule()
molecule.add_atom(Atom(symbol='O', coords=(0,0,0)))
molecule.add_atom(Atom(symbol='H', coords=(1,0,0)))
molecule.add_atom(Atom(symbol='H', coords=(0,1,0)))

settings = Settings()
settings.input.ams.Task = 'GeometryOptimization'
settings.input.ams.Properties.NormalModes = 'Yes'
settings.input.dftb.Model = 'GFN1-xTB'

job = AMSJob(molecule=molecule, settings=settings, name='water_optimization')
result = job.run()

energy = result.get_energy(unit='kcal/mol')
frequencies = result.get_frequencies(unit='cm^-1')
optimized_molecule = result.get_main_molecule()

# Unlike python lists, where the index of the first element is 0, 
# the index of the first atom in the molecule object is 1
bond_angle = optimized_molecule[1].angle(optimized_molecule[2], optimized_molecule[3])

print('== Results ==')
print('Optimized geometry:')
print(optimized_molecule)
print('Energy      : {:.3f} kcal/mol'.format(energy))
print('Bond angle  : {:.1f} degrees'.format(Units.convert(bond_angle, 'rad', 'degree')))
print('Frequencies : {} cm^-1'.format(frequencies))


from collections import OrderedDict
from itertools import product
import numpy as np

from ..core.results import Results
from ..core.basejob import MultiJob
from ..tools.units import Units

__all__ = ['NumHessJob', 'NumHessResults'] #names exported to the main namespace

class NumHessResults(Results):

    def get_hessian(self, mass_weighted=False):
        j = self.job
        n = len(j.molecule)
        hessian = np.empty((3*n,3*n))

        for (atom,axis) in product(range(1,n+1), range(3)):
            v1 = np.array(j.gradient(j.children[(atom, axis, -1)].results))
            v2 = np.array(j.gradient(j.children[(atom, axis, 1)].results))
            hessian[3*(atom-1) + axis] = (v2 - v1)/(2 * Units.convert(j.step,'angstrom', 'bohr'))

        if mass_weighted:
            masses = [at.mass for at in j.molecule]
            weights = np.empty((n,n))
            for i,j in product(range(n), repeat=2):
                weights[i,j] = masses[i] * masses[j]
            hessian /= np.kron(weights, np.ones((3,3)))
        return (hessian+hessian.T)/2


class NumHessJob(MultiJob):
    """A class for calculating numerical hessian.

    The length and unit of the geometry displacement step can be adjusted.
    *gradient* should be a function that takes results of *jobtype* and
    returns energy gradient in hartree/bohr.
    """
    _result_type = NumHessResults

    def __init__(self, molecule, step=0.01, unit='angstrom', jobtype=None, gradient=None, **kwargs):
        MultiJob.__init__(self, children=OrderedDict(), **kwargs)
        self.molecule = molecule
        self.step = Units.convert(step, unit, 'angstrom')
        self.jobtype = jobtype   #who is going to calculate single points
        self.gradient = gradient   #function extracting gradients from children

    def prerun(self):
        for (atom,axis,step) in product(range(1,1+len(self.molecule)), range(3), [-1,1]):
            vec = [0,0,0]
            vec[axis] = self.step * step
            newmol = self.molecule.copy()
            newmol[atom].translate(vec)
            newname = '{}_{}_{}'.format(atom,axis,step)
            self.children[(atom,axis,step)] = self.jobtype(name=newname, molecule=newmol,
                                                           settings=self.settings)



from ...core.functions import add_to_class
from ...mol.molecule import Molecule
from ...mol.atom import Atom
from numpy import zeros as npz
from numpy import array as npa

__all__ = ['toASE', 'fromASE']
ase_present = False

try:
    from ase import Atom as aseAtom
    from ase import Atoms as aseAtoms
    ase_present = True
except ImportError:
    __all__ = []


@add_to_class(Molecule)
def readase(self, f, **other):
    """Read Molecule using ASE engine

    The ``read`` function of the |Molecule| class passes a file descriptor into here, so in this case you must specify the *format* to be read by ASE::

        mol = Molecule('file.cif', inputformat='ase', format='cif')

    The ASE Atoms object then gets converted to a PLAMS Molecule and returned.
    All *other* options are passed to ``ASE.io.read()``.
    See https://wiki.fysik.dtu.dk/ase/ase/io/io.html on how to use it.

    .. note::

        The nomenclature of PLAMS and ASE is incompatible for reading multiple geometries, make sure that you only read single geometries with ASE! Reading multiple geometries is not supported, each geometry needs to be read individually.

    """
    try:
        from ase import io as aseIO
    except ImportError:
        raise MoleculeError('Asked for ASE IO engine but could not load ASE.io module')

    aseMol = aseIO.read(f, **other)
    mol = fromASE(aseMol)
    #update self with the molecule read without overwriting e.g. settings
    self += mol
    #lattice does not survive soft update
    self.lattice = mol.lattice
    return


@add_to_class(Molecule)
def writease(self, f, **other):
    """Write molecular coordinates using ASE engine.

    The ``write`` function of the |Molecule| class passes a file descriptor into here, so in this case you must specify the *format* to be written by ASE.
    All *other* options are passed to ``ASE.io.write()``.
    See https://wiki.fysik.dtu.dk/ase/ase/io/io.html on how to use it.

    These two write the same content to the respective files::

        molecule.write('filename.anyextension', outputformat='ase', format='gen')
        molecule.writease('filename.anyextension', format='gen')

    """
    aseMol = toASE(self)
    aseMol.write(f, **other)
    return

if ase_present:
    Molecule._readformat['ase'] = Molecule.readase
    Molecule._writeformat['ase'] = Molecule.writease



def toASE(molecule):
    """Convert a PLAMS |Molecule| to an ASE molecule (``ase.Atoms`` instance). Translate coordinates, atomic numbers, and lattice vectors (if present). The order of atoms is preserved."""
    aseMol = aseAtoms()

    #iterate over PLAMS atoms
    for atom in molecule:

        #check if coords only consists of floats or ints
        if not all(isinstance(x, (int,float)) for x in atom.coords):
            raise ValueError("Non-Number in Atomic Coordinates, not compatible with ASE")

        #append atom to aseMol
        aseMol.append(aseAtom(atom.atnum, atom.coords))

    #get lattice info if any
    lattice = npz((3,3))
    pbc = [False,False,False]
    for i,vec in enumerate(molecule.lattice):

        #check if lattice only consists of floats or ints
        if not all(isinstance(x, (int,float)) for x in vec):
            raise ValueError("Non-Number in Lattice Vectors, not compatible with ASE")

        pbc[i] = True
        lattice[i] = npa(vec)

    #save lattice info to aseMol
    if any(pbc):
        aseMol.set_pbc(pbc)
        aseMol.set_cell(lattice)

    return aseMol



def fromASE(molecule):
    """Convert an ASE molecule to a PLAMS |Molecule|. Translate coordinates, atomic numbers, and lattice vectors (if present). The order of atoms is preserved."""
    plamsMol = Molecule()

    #iterate over ASE atoms
    for atom in molecule:
        #add atom to plamsMol
        plamsMol.add_atom(Atom(atnum=atom.number, coords=tuple(atom.position)))

    #add Lattice if any
    if any(molecule.get_pbc()):
        lattice = []
        #loop over three booleans
        for i,boolean in enumerate(molecule.get_pbc().tolist()):
            if boolean:
                lattice.append(tuple(molecule.get_cell()[i]))

        #write lattice to plamsMol
        plamsMol.lattice = lattice.copy()


    return plamsMol



from collections import OrderedDict
from itertools import product

from ..core.results import Results
from ..core.basejob import MultiJob

__all__ = ['NumGradJob', 'NumGradResults'] #names exported to the main namespace

class NumGradResults(Results):

    def _get_gradient_component(self, atom, coord, extract):
        """Get gradient component for *atom* along *coord*.

        *atom* should be integer, *coord* a single character string with 'x', 'y' or 'z'.
        *extract* should be a function that takes results of a single point job and
        returns a single number.
        """
        energies = [extract(self.job.children[(atom,coord,i)].results) for i in self.job.steps]
        coeffs, denom = self.job._coeffs[self.job.npoints]
        return sum([c*e for c,e in zip(coeffs, energies)])/(denom * self.job.step)

    def get_gradient(self, extract):
        """Get the full gradient vector. Returns a list of length 3N,
        where N is the number of atoms in the calculated molecule.

        *extract* should be a function that takes results of a single point job and
        returns a single number.
        """
        return [self._get_gradient_component(atom, coord, extract)
            for atom,coord in product(range(1,1+len(self.job.molecule)), 'xyz')]


class NumGradJob(MultiJob):
    """A class for calculating numerical gradients of energy
    (or some other single-valued property) with respect to
    cartesian displacements.

    The differentiation is done using the final difference method
    with 2, 4, 6 or 8 points. The length of the step can be adjusted.
    """
    _result_type = NumGradResults

    #finite difference coefficients for different number of points:
    _coeffs = {
        2: ([-1,1], 2),
        4: ([1,-8,8,-1], 12),
        6: ([-1,9,-75,75,-9,1], 60),
        8: ([3,-32,168,-672,672,-168,32,-3], 840),
    }

    def __init__(self, molecule, npoints=2, step=0.01, unit='angstrom', jobtype=None, **kwargs):
        #initialize parent class and store all additional constructor arguments
        MultiJob.__init__(self, children=OrderedDict(), **kwargs)
        self.molecule = molecule
        self.npoints = npoints
        self.step = step
        self.unit = unit
        self.jobtype = jobtype   #who is going to calculate single points


    def prerun(self):
        self.steps = list(range(-(self.npoints//2), 0)) + list(range(1, self.npoints//2+1))

        for (atom,axis,i) in product(range(1,1+len(self.molecule)), 'xyz', self.steps):
            v = (self.step * i if axis == 'x' else 0,
                 self.step * i if axis == 'y' else 0,
                 self.step * i if axis == 'z' else 0)
            newmol = self.molecule.copy()
            newmol[atom].translate(v, self.unit)
            newname = '{}_{}_{}'.format(atom,axis,i)
            #settings of NumGradJob are directly passed to children single point jobs
            self.children[(atom,axis,i)] = self.jobtype(name=newname, molecule=newmol,
                                                        settings=self.settings)




from .scmjob import SCMJob, SCMResults
from ...core.errors import ResultsError
from ...core.functions import log
from ...tools.units import Units

__all__ = ['BANDJob', 'BANDResults']


class BANDResults(SCMResults):
    _kfext = '.runkf'
    _rename_map = {'RUNKF':'$JN'+_kfext}


    def get_properties(self):
        """get_properties()
        Return a dictionary with all the entries from ``Properties`` section in the main KF file (``$JN.rkf``).
        """
        n = self.readkf('Properties', 'nEntries')
        ret = {}
        for i in range(1, n+1):
            tp = self.readkf('Properties', 'Type({})'.format(i)).strip()
            stp = self.readkf('Properties', 'Subtype({})'.format(i)).strip()
            val = self.readkf('Properties', 'Value({})'.format(i))
            if stp.endswith(' (a.u.)'):
                stp = stp[:-7]
            key = '{} {}'.format(stp, tp.lower())
            ret[key] = val
        return ret


    def get_main_molecule(self):
        """get_main_molecule()
        Return a |Molecule| instance based on the ``Molecule`` section in the main KF file (``$JN.runkf``).

        For runs with multiple geometries (geometry optimization, transition state search, molecular dynamics) this is the **final** geometry. In such a case, to access the initial (or any intermediate) coordinates please use :meth:`get_input_molecule` or extract coordinates from section ``History``, variables ``xyz0``, ``xyz1`` and so on. Mind the fact that all coordinates written by BAND to ``History`` section are in bohr and internal atom order::

            mol = results.get_molecule(section='History', variable='xyz0', unit='bohr', internal=True)
        """
        ret = self.get_molecule(section='Molecule', variable='Coords', unit='bohr', internal=True)
        ret.properties.charge = self.readkf('Molecule', 'Charge')
        if ('Molecule', 'LatticeVectors') in self._kf:
            lattice = self.readkf('Molecule', 'LatticeVectors')
            lattice = Units.convert(lattice, 'bohr', 'angstrom')
            ret.lattice = [tuple(lattice[i:i+3]) for i in range(0,len(lattice),3)]
        return ret


    def get_input_molecule(self):
        """get_input_molecule()
        Return a |Molecule| instance with initial coordinates.

        All data used by this method is taken from ``$JN.runkf`` file. The ``molecule`` attribute of the corresponding job is ignored.
        """
        if ('History', 'nr of geometries') in self._kf:
            ret = self.get_molecule(section='History', variable='xyz0', unit='bohr', internal=False)
            lattice = self.readkf('History', 'lattice0')
            lattice = Units.convert(lattice, 'bohr', 'angstrom')
            lattice = [lattice[i:i+3] for i in range(0,len(lattice),3)]
            while len(lattice) > 0 and not any(lattice[-1]):
                lattice.pop()
            ret.lattice = [tuple(i) for i in lattice]
            return ret
        return self.get_main_molecule()


    def get_energy(self, unit='au'):
        """get_energy(unit='au')
        Return final bond energy, expressed in *unit*.
        """
        return self._get_single_value('Bond energies', 'final bond energy', unit)


    def get_fermi_energy(self, unit='au'):
        """get_fermi_energy(unit='au')
        Return Fermi energy, expressed in *unit*.
        """
        return self._get_single_value('BandStructure', 'FermiEnergy', unit)


    def get_band_gap(self, unit='au'):
        """get_band_gap(unit='au')
        Return the value of the band gap, expressed in *unit*.
        """
        return self._get_single_value('BandStructure', 'BandGap', unit)


    def get_dipole_vector(self, unit='au'):
        """get_dipole_vector(unit='au')
        Return the dipole vector, expressed in *unit*.
        """
        prop = self.get_properties()
        if 'Dipole moment vector' in prop:
            return Units.convert(prop['Dipole moment vector'], 'au', unit)
        raise ResultsError("'Dipole moment vector' not present in 'Properties' section of {}".format(self._kfpath()))


    def _int2inp(self):
        """_int2inp()
        Get mapping from the internal atom order to the input atom order.
        """
        return self.readkf('geometry', 'Atom map new order')


    def _atomic_numbers_input_order(self):
        """_atomic_numbers_input_order()
        Return a list of atomic numbers, in the input order.
        """
        mapping = self._int2inp()
        atnums = self.readkf('Molecule', 'AtomicNumbers')
        return [atnums[mapping[i]-1] for i in range(len(atnums))]



class BANDJob(SCMJob):
    _result_type = BANDResults
    _command = 'band'


    def _serialize_mol(self):
        s = self.settings.input
        s.units.length = 'angstrom'

        if len(self.molecule.lattice) in [1,2] and self.molecule.align_lattice():
            log("The lattice supplied for job {} did not follow the convention required by BAND. I rotated the whole system for you. You're welcome".format(self._full_name()), 3)

        for i,atom in enumerate(self.molecule):
            s.atoms['_'+str(i+1)] = atom.str(symbol=self._atom_symbol(atom), space=18, decimal=10)

        if self.molecule.lattice:
            for i,vec in enumerate(self.molecule.lattice):
                s.lattice['_'+str(i+1)] = '{:16.10f} {:16.10f} {:16.10f}'.format(*vec)


    def _remove_mol(self):
        if 'atoms' in self.settings.input:
            del self.settings.input.atoms
        if 'lattice' in self.settings.input:
            del self.settings.input.lattice


from ...core.errors import PlamsError
from .scmjob import SCMJob, SCMResults

__all__ = ['AMSAnalysisJob', 'AMSAnalysisResults','convert_to_unicode']

class AMSAnalysisPlot:
    """
    Class representing a plot of 2D or higher

    * ``x``       -- A list of lists containing the values in each of the multiple x-axes
    * ``y``       -- A list containing the values along the y-axis
    * ``y_sigma`` -- A list containing the standard deviation of the values onthe y-axis
    * ``name``    -- The name of the plot

    The most important method is the write method, which returns a string containing all the plot info,
    and can also write a corresponding file if a filename is provided as argument.
    This file can be read by e.g. gnuplot.
    """
    def __init__(self):
        """
        Initiate an instance of the plot class
        """
        self.x = []
        self.x_units = []
        self.x_names = []

        self.y = None
        self.y_units = None
        self.y_name = None
        self.y_sigma = None # stadard deviation for y_values

        self.properties = None
        self.name = None

    def read_data (self, kf, sec) :
        """
        Read the xy data for a section from the kf file
        """
        # Read all the x-values. There can be multiple axes for ND plots (n=3,4,....)
        xkeys = [k for k in kf.reader._sections[sec] if 'x(' in k and ')-axis' in k]
        xnums = sorted([int(k.split('(')[1].split(')')[0]) for k in xkeys])
        for i in xnums :
                xkey = 'x(%i)-axis'%(i)
        self.x.append(kf.read(sec, xkey))
        x_name = kf.read(sec, '%s(label)'%(xkey))
        self.x_names.append(convert_to_unicode(x_name))
        self.x_units.append(convert_to_unicode(kf.read(sec, '%s(units)'%(xkey))))

        # Read the y-values
        ykey = 'y-axis'
        y_name = kf.read(sec, '%s(label)'%(ykey))
        self.y = kf.read(sec, ykey)
        self.y_name = convert_to_unicode(y_name)
        self.y_units = convert_to_unicode(kf.read(sec, '%s(units)'%(ykey)))

        self.y_sigma = kf.read(sec, 'sigma')

        self.read_properties(kf, sec)

    def read_properties (self, kf, sec) :
        """
        Read properties from the KF file
        """
        counter = 0
        properties = {}
        while(1) :
            counter += 1
            try :
                propname = kf.read(sec, 'Property(%i)'%(counter)).strip()
            except :
                break
            properties[propname] = kf.read(sec, propname)
            if isinstance(properties[propname],str) :
                properties[propname] = properties[propname].strip()

        # Now set the instance variables
        self.properties = properties
        if 'Legend' in properties :
            self.name = properties['Legend']

    def write (self, outfilename=None) :
        """
        Print this plot to a text file
        """
        # Place property string
        parts = []
        for propname,prop in self.properties.items() :
            parts.append('%-30s %s\n'%(propname, prop))

        # Place the string with the column names
        x_name = ''
        for xname,xunit in zip(self.x_names,self.x_units) :
            x_str = '%s(%s)'%(xname,self.xunit)
            x_name += '%30s '%(x_str)
        y_name = '%s(%s)'%(self.y_name,self.y_units)
        parts.append('%s %30s %30s\n'%(x_name,y_name,'sigma'))

        # Place the values
        value_lists = self.x + [self.y] + [self.y_sigma]
        for values in zip(*value_lists) :
            v_str = ''
            for v in values :
                v_str += '%30.10e '%(v)
            v_str += '\n'
            parts.append(v_str)
        block = ''.join(parts)

        if outfilename is not None :
            outfile = open(outfilename,'w',encoding='utf8')
            outfile.write(block)
            outfile.close()

        return block
        

class AMSAnalysisResults(SCMResults):
    _kfext = '.kf'
    _rename_map = {'plot.kf':'$JN'+_kfext}

    def get_molecule(self, *args, **kwargs):
        raise PlamsError('AMSAnalysisResults does not support the get_molecule() method.')

    def get_sections(self) :
        """
        Read the sections available to make xy plots
        """
        if not self._kfpresent():
            raise FileError('File {} not present in {}'.format(self.job.name+self.__class__._kfext, self.job.path))
        if self._kf.reader._sections is None :
            self._kf.reader._create_index()
        sections = self._kf.reader._sections.keys()
        return sections

    def get_xy(self, section='', i=1):
        xy = AMSAnalysisPlot()

        task = self.job.settings.input.Task
        if section == '' :
            section = task

        # Find the correct section in the KF file
        sections = self.get_sections()
        matches = [s for s in sections if s.lower()==section.lower()+'(%i)'%(i)]
        if len(matches) == 0 :
                print ('Sections: ',list(sections))
                raise PlamsError('AMSAnalysisResults.get_xy(section,i): section must be one of the above. You specified "{}"'.format(section))
        sec = matches[0] 

        # Get the data
        xy.read_data(self._kf,sec)

        return xy

    def get_all_plots (self) :
        """
        Get a list of all the plot objects created by the analysis jobs
        """
        sections = self.get_sections()
        plots = []
        for section in sections :
            name_part = section.split('(')[0]
            num_part = int(section.split('(')[1].split(')')[0])
            outfilename = '%s_%i.dat'%(name_part,num_part)
            xy = self.get_xy(name_part,num_part)
            plots.append(xy)
        return plots

    def write_all_plots (self) :
        """
        Write all the plots created by the analysis job to file
        """
        plots = self.get_all_plots()
        for xy in plots :
            xy.write('%s'%(outfilename))

    def get_D(self, i=1):
        """ returns a 2-tuple (D, D_units) from the AutoCorrelation(i) section on the .kf file. """

        # If there are multiple, it will read the first one
        sections = [sec for sec in self.get_sections() if 'Integral' in sec]
        if len(sections) < i : 
            return None,None
        section = sections[i-1]
        plot = self.get_xy(section.split('(')[0],i)
        if not 'DiffusionCoefficient' in plot.properties.keys() :
            return None, None

        D = plot.properties['DiffusionCoefficient']
        D_units = plot.y_units
        return D, D_units


class AMSAnalysisJob(SCMJob):
    """A class for analyzing molecular dynamics trajectories using the ``analysis`` program.


    """
    _result_type = AMSAnalysisResults
    _command = 'analysis'
    _subblock_end = 'end'

    def __init__(self, **kwargs):
        SCMJob.__init__(self, **kwargs)

    def _serialize_mol(self):
        pass

    def _remove_mol(self):
        pass

    def check(self):
        try:
            grep = self.results.grep_file('$JN.err', 'NORMAL TERMINATION')
        except:
            return False
        return len(grep) > 0

def convert_to_unicode (k) :
    """
    Convert a string with ascii symbols representing unicode symbols

    Example k: 'abc\\u03c9def'
    """
    parts = k.split('\\u')
    # Collect the hexadecimals
    symbols = [chr(int(part[:4],16)) for part in parts[1:]]
    # Now repair the parts
    parts = parts[:1] + [''.join([s,part[4:]]) for s,part in zip(symbols,parts[1:])]
    key = ''.join(parts)

    return key



import os
import sys
import copy
import hashlib
import subprocess
import time
import warnings
from typing import Callable, Dict, NoReturn

from os.path import join as opj
from contextlib import AbstractContextManager


from .functions import config, log


__all__ = []



def smart_copy(obj, owncopy=[], without=[]):
    """Return a copy of *obj*. Attributes of *obj* listed in *without* are ignored. Attributes listed in *owncopy* are copied by calling their own ``copy()`` methods. All other attributes are copied using :func:`copy.deepcopy`."""

    ret = obj.__class__()
    for k in owncopy:
        ret.__dict__[k] = obj.__dict__[k].copy()
    for k in obj.__dict__:
        if k not in (without + owncopy):
            ret.__dict__[k] = copy.deepcopy(obj.__dict__[k])
    return ret


#===========================================================================


def sha256(string):
    """A small utility wrapper around :ref:`hashlib.sha256<hash-algorithms>`."""
    if not isinstance(string, bytes):
        string = str(string).encode()
    h = hashlib.sha256()
    h.update(string)
    return h.hexdigest()


#===========================================================================


def saferun(*args, **kwargs):
    """A wrapper around :func:`subprocess.run` repeating the call ``config.saferun.repeat`` times with ``config.saferun.delay`` interval in case of :exc:`BlockingIOError` being raised (any other exception is not caught and directly passed above). All arguments (*args* and *kwargs*) are passed directly to :func:`~subprocess.run`. If all attempts fail, the last raised :exc:`BlockingIOError` is reraised."""
    attempt = 0
    (repeat, delay) = (config.saferun.repeat, config.saferun.delay) if ('saferun' in config) else (5,1)
    while attempt <= repeat:
        try:
            return subprocess.run(*args, **kwargs)
        except OSError as e:
            attempt += 1
            log('subprocess.run({}) attempt {} failed with {}'.format(args[0], attempt, e), 5)
            last_error = e
            time.sleep(delay)
    raise last_error


#===========================================================================


class UpdateSysPath(AbstractContextManager):
    """A context manager for temporary adding ``$AMSHOME/scripting`` to :data:`sys.path`.

    While the context manager is opened modules can be imported (directly) from the ``$AMSHOME/scripting`` directory,
    allowing one access to the python modules distributed with ADF.
    An :exc:`EnvironmentError` is raised if the ``AMSHOME`` environment variable has not been set.

    Usage example:

    .. code:: python
        >>> import sys
        >>> import os

        >>> scripting = os.path.join(os.environ['AMSHOME'], 'scripting')
        >>> with UpdateSysPath():
        ...     print(scripting in sys.path)  # The context manager is opened
        True

        >>> print(scripting in sys.path)  # The context manager is closed
        False

    """

    def __init__(self, path: str = None) -> None:
        try:
            parser_path = path if path is not None else opj(os.environ['AMSHOME'], 'scripting')
        except KeyError as ex:
            raise EnvironmentError("The 'AMSHOME' environment variable has not been set").with_traceback(ex.__traceback__)

        if parser_path not in sys.path:
            sys.path.append(parser_path)
            self.path = parser_path
        else:  # The specified path is already present in sys.path
            self.path = None

    def __enter__(self) -> None:
        """If not ``None``, add :attr:`UpdateSyspath.path` to :data:`sys.path`."""
        if self.path is not None:
            sys.path.append(self.path)

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        """If not ``None``, remove :attr:`UpdateSyspath.path` from :data:`sys.path`."""
        if self.path is not None:
            try:
                idx = sys.path.index(self.path)
            except ValueError:  # self.path has been (manually) removed by the user
                pass
            else:
                del sys.path[idx]

#===========================================================================

def _raise(exc: Exception) -> NoReturn:
    """Raise *exc*; this tiny function is here because ``lambda`` expressions can't handle the ``raise`` statement."""
    raise exc


#: Map a :class:`str` representation of an action to an actual callable.
_ACTION_DICT: Dict[str, Callable[[Exception], None]] = {
    'ignore': lambda n: None,
    'raise': lambda n: _raise(n),
    'warn': lambda n: warnings.warn(str(n), stacklevel=2)
}


def parse_action(action: str) -> Callable[[Exception], None]:
    """Return a callable which takes an :exc:`Exception` as argument and performs some action base on the value of *action*.

    * ``action = 'ignore'``: Do nothing.
    * ``action = 'warn'``: Issue a warning.
    * ``action = 'raise'``: Raise the passed exception.

    """
    try:
        return _ACTION_DICT[action]
    except KeyError as ex:
        raise ValueError("`action` expected 'ignore', 'raise' or 'warn'; "
                         f"observed value: {action!r}") from ex


import multiprocessing

# Run jobs as many jobs as possible in parallel:
config.default_jobrunner = JobRunner(parallel=True, maxjobs=multiprocessing.cpu_count())
config.job.runscript.nproc = 1

# The molecules we want to use in our benchmark:
mol_smiles = {'Methane'  : 'C',
              'Ethane'   : 'C-C',
              'Ethylene' : 'C=C',
              'Acetylene': 'C#C'}
molecules = {}
for name, smiles in mol_smiles.items():
    # Compute 10 conformers, optimize with UFF and pick the lowest in energy.
    molecules[name] = from_smiles(smiles, nconfs=10, forcefield='uff')[0]
    print(name, molecules[name])

# Initialize the common settings:
common_settings = Settings()
common_settings.input.ams.Task = 'SinglePoint'
common_settings.input.ams.System.Symmetrize = 'Yes'
common_settings.input.adf.Basis.Core = 'None'

basis = ['QZ4P', 'TZ2P', 'TZP', 'DZP', 'DZ', 'SZ']
reference_basis = 'QZ4P'

# Set up and run the calculations: 
results = {}
for bas in basis:
    for name, molecule in molecules.items():
        settings = common_settings.copy()
        settings.input.adf.Basis.Type = bas
        job = AMSJob(name=name+'_'+bas, molecule=molecule, settings=settings)
        results[(name,bas)] = job.run()

# Calculate the average absolute error in bond energy per atom for each basis set:
average_errors = {}
for bas in basis:
    if bas != reference_basis:
        errors = []
        for name, molecule in molecules.items():
            reference_energy = results[(name,reference_basis)].get_energy(unit='kcal/mol')
            energy = results[(name,bas)].get_energy(unit='kcal/mol')
            errors.append(abs(energy - reference_energy)/len(molecule))
            print('Energy for {} using {} basis set: {} [kcal/mol]'.format(name,bas,energy))
        average_errors[bas] = sum(errors)/len(errors)

print('== Results ==')
print('Average absolute error in bond energy per atom')
for bas in basis:
    if bas != reference_basis:
        print('Error for basis set {:<4}: {:>10.3f} [kcal/mol]'.format(bas, average_errors[bas]))



# Compute the neutral-anion reorganization energy of pyrrole 
# using ADF as computational engine

molecule = Molecule('pyrrole.xyz')

# Generic settings of the calculation
# (for quantitatively better results, use better settings)
common_settings = Settings()
common_settings.input.adf.Basis.Type = 'DZ'

# Specific settings for the neutral calculation.
# Nothing special needs to be done for the neutral calculation,
# so we just use an empty settings.
neutral_settings = Settings()

# Specific settings for the anion calculation:
anion_settings = Settings()
anion_settings.input.ams.System.Charge = -1
anion_settings.input.adf.Unrestricted = 'Yes'
anion_settings.input.adf.SpinPolarization = 1

# Create and run the ReorganizationEnergyJob:
job = ReorganizationEnergyJob(molecule, common_settings, neutral_settings,
                              anion_settings, name=molecule.properties.name)
job.run()

# Fetch and print the results:

energy_unit = 'eV'
energies = job.results.get_all_energies(energy_unit)
reorganization_energy = job.results.reorganization_energy(energy_unit)

print('')
print("== Results ==")
print('')
print(f"Molecule: {molecule.properties.name}")
print("State A: neutral")
print("State B: anion")
print('')
print(f'Reorganization energy: {reorganization_energy:.6f} [{energy_unit}]')
print('')
print(f'|   State   | Optim Geo | Energy [{energy_unit}]')
print(f'|     A     |     A     | {energies["state A geo A"]:.6f}')
print(f'|     A     |     B     | {energies["state A geo B"]:.6f}')
print(f'|     B     |     A     | {energies["state B geo A"]:.6f}')
print(f'|     B     |     B     | {energies["state B geo B"]:.6f}')
print('')



import os
import functools
import threading
import time

from os.path import join as opj
from subprocess import DEVNULL, PIPE

from .errors import PlamsError
from .functions import config, log
from .private import saferun
from .settings import Settings


__all__ = ['JobRunner', 'GridRunner']



def _in_thread(func):
    """Decorator for an instance method. If ``parallel`` attribute of given instance is ``True``, run decorated method in a separate :class:`~threading.Thread`. This thread is usually a daemon thread, the decision is based on ``config.daemon_threads`` entry."""
    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
        if self.parallel:
            t = threading.Thread(name='plamsthread', target=func, args=(self,)+args, kwargs=kwargs)
            t.daemon = config.daemon_threads
            t.start()
        else:
            func(self, *args, **kwargs)
    return wrapper


def _limit(func):
    """Decorator for an instance method. If ``semaphore`` attribute of given instance is not ``None``, use this attribute to wrap decorated method via :ref:`with<with-locks>` statement."""
    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
        if self.semaphore:
            with self.semaphore:
                return func(self, *args, **kwargs)
        else:
            return func(self, *args, **kwargs)
    return wrapper


class _MetaRunner(type):
    """Metaclass for |JobRunner|. During an instance creation wrap the :meth:`~scm.plams.core.jobrunner.JobRunner.call` method with :func:`_limit` decorator which enforces a limit on the number of simultaneous :meth:`~scm.plams.core.jobrunner.JobRunner.call` calls.
    """
    def __new__(meta, name, bases, dct):
        dct['call'] = _limit(dct['call'])
        return type.__new__(meta, name, bases, dct)



#===========================================================================
#===========================================================================
#===========================================================================



class JobRunner(metaclass=_MetaRunner):
    """Class defining the basic job runner interface. Instances of this class represent local job runners -- job runners that execute computational jobs on the current machine.

    The goal of the job runner is to take care of two important things -- parallelization and runscript execution:

    *   When the method |run| of any |Job| instance is executed, the control, after some initial preparations, is passed to a |JobRunner| instance. This |JobRunner| instance decides if a separate thread should be spawned for the job or if the execution should proceed in the current thread. This decision is based on the ``parallel`` attribute which can be set on |JobRunner| creation. There are no separate classes for serial and parallel job runner, both cases are covered by |JobRunner| depending on the ``parallel`` parameter.
    *   If the executed job is an instance of |SingleJob|, it creates a runscript which contains most of the actual computational work (usually it's just an execution of some external binary). The runscript is then submitted to a |JobRunner| instance using its :meth:`call` method. This method executes the runscript as a separate subprocess and takes care of output and error streams handling, setting a proper working directory etc.

    For a job runner with parallel execution enabled the number of simultaneously running jobs can be limited using the *maxjobs* parameter. If *maxjobs* is 0, no limit is enforced. If *parallel* is ``False``, *maxjobs* is ignored. If *parallel* is ``True`` and *maxjobs* is a positive integer, a :class:`BoundedSemaphore<threading.BoundedSemaphore>` of that size is used to limit the number of simultaneously running :meth:`call` methods.

    A |JobRunner| instance can be passed to |run| with a keyword argument ``jobrunner``. If this argument is omitted, the instance stored in ``config.default_jobrunner`` is used.
    """

    def __init__ (self, parallel=False, maxjobs=0):
        self.parallel = parallel
        self.maxjobs  = maxjobs
        self.semaphore = threading.BoundedSemaphore(maxjobs) if maxjobs else None


    def call(self, runscript, workdir, out, err, runflags):
        """call(runscript, workdir, out, err, runflags)
        Execute the *runscript* in the folder *workdir*. Redirect output and error streams to *out* and *err*, respectively.

        Arguments *runscript*, *workdir*, *out* and *err* should be strings with paths to corresponding files or folders.

        *runflags* is a |Settings| instance containing the `run` branch of running job's settings. The basic job runner defined here ignores them, but they can be useful in |JobRunner| subclasses (see :meth:`GridRunner.call`).

        Returns an integer with the exit code returned by the *runscript*.

        This method can be safely overridden in |JobRunner| subclasses. For example, in |GridRunner| it submits the runscript to a queueing system instead of executing it locally.

        .. note::
            This method is used automatically during |run| and should never be explicitly called in your script.
        """
        log('Executing {}'.format(runscript), 5)
        command = ['./'+runscript] if os.name == 'posix' else ['sh', runscript]
        if out is not None:
            with open(opj(workdir, err), 'w') as e, open(opj(workdir, out), 'w') as o:
                process = saferun(command, cwd=workdir, stderr=e, stdout=o)
        else:
            with open(opj(workdir, err), 'w') as e:
                process = saferun(command, cwd=workdir, stderr=e)
        log('Execution of {} finished with returncode {}'.format(runscript, process.returncode), 5)
        return process.returncode


    @_in_thread
    def _run_job(self, job, jobmanager):
        """_run_job(job, jobmanager)
        This method aggregates the parts of :ref:`job-life-cycle` that are supposed to be run in a separate thread in case of parallel job execution. It is wrapped with :func:`_in_thread` decorator.

        This method should not be overridden.
        """
        if job._prepare(jobmanager):
            job._execute(self)
            job._finalize()



#===========================================================================
#===========================================================================
#===========================================================================



class GridRunner(JobRunner):
    """Subclass of |JobRunner| that submits the runscript to a queueing system instead of executing it locally. Besides two new keyword arguments (*grid* and *sleepstep*) it behaves and is meant to be used just like a regular |JobRunner|.

    .. note::

        The default value of the *parallel* argument is ``True``, unlike in the regular |JobRunner|.

    There are many different queueing systems that are popular nowadays (for example: TORQUE, SLURM, OGE). Usually they use different commands for submitting jobs or checking the queue status. |GridRunner| class tries to build a common interface to these systems. The commands used to communicate with the queueing system are not hard-coded, but rather taken from a |Settings| instance. Thanks to that the user has almost full control over the behavior of a |GridRunner| instance and the behavior can be ajdusted dynamically.

    The behavior of a |GridRunner| instance is determined by the contents of a |Settings| instance stored in its ``settings`` attribute. That |Settings| instance can be manually supplied by the user or taken from a collection of predefined instances stored as branches of ``GridRunner.config``. The adjustment is done with the *grid* parameter which should be a string or a |Settings| instance. If it's a string, it has to be a key occurring in ``GridRunner.config`` or ``'auto'`` for autodetection. For example, if ``grid='slurm'`` is passed, ``GridRunner.config.slurm`` is used as settings. If ``grid='auto'`` then entries present in ``GridRunner.config`` are tested and the first one that works (its submit command is present on your system) is chosen. When a |Settings| instance is passed as *grid*, it is directly used as ``settings``.

    Currently two predefined schemes are available (see :ref:`plams-defaults`): ``slurm`` for SLURM and ``pbs`` for queueing systems following PBS syntax (PBS, TORQUE, Oracle Grid Engine etc.).

    The settings of |GridRunner| should have the following structure:

    *   ``output`` -- flag for specifying the output file path.
    *   ``error`` -- flag for specifying the error file path.
    *   ``workdir`` -- flag for specifying path to the working directory.
    *   ``commands.submit`` -- submit command.
    *   ``commands.check`` -- queue status check command.
    *   ``commands.getid`` -- function extracting submitted job's ID from the output of the submit command.
    *   ``commands.running`` -- function extracting a list of all running jobs from the output of queue check command
    *   ``commands.special`` -- branch storing definitions of special |run| keyword arguments.

    See :meth:`call` for more details and examples.

    The *sleepstep* parameter defines how often the queue check is performed. It should be a numerical value telling how many seconds should the interval between two consecutive checks last.

    .. note::
        Usually queueing systems are configured in such a way that output of your calculation is captured somewhere else and copied to the location indicated by the output flag only when the job is finished. Because of that it is not possible to have a peek at your output while your job is running (for example, to see if your calculation is going well). This limitation can be circumvented with ``myjob.settings.runscript.stdout_redirect`` flag. If set to ``True``, the output redirection will not be handled by the queueing system, but rather placed in the runscript using the shell redirection ``>``. That forces the output file to be created directly in *workdir* and updated live as the job proceeds.
    """

    # GridRunner mechanism for testing if a job is finished:
    #if [...].commands.finished exists it is used to check if the job is finished. It should be a function that takes a single string (job_id) as an argument and returns True or False
    #otherwise [...].commands.check is combined with job_id, executed as a subprocess and returned exit code is tested (nonzero return code indicates that job has finished)

    def __slurm_get_jobid(output):
        s = output.split()
        if len(s) > 0 and all([ch.isdigit() for ch in s[-1]]):
            return s[-1]
        return None

    def __slurm_running(output):
        lines = output.splitlines()[1:]
        return [line.split()[0] for line in lines]

    def __pbs_get_jobid(output):
        s = output.split('.')
        if len(s) > 0 and all([ch.isdigit() for ch in s[0]]):
            return s[0]
        return None

    def __pbs_running(output):
        lines = output.splitlines()[2:]
        return [line.split()[0].split('.')[0] for line in lines]

    # Static config holding the preconfigured grids:
    config = Settings()
    # PBS
    config.pbs.workdir = '-d'
    config.pbs.output  = '-o'
    config.pbs.error   = '-e'
    config.pbs.special.nodes    = '-l nodes='
    config.pbs.special.walltime = '-l walltime='
    config.pbs.special.memory = '-l mem='
    config.pbs.special.queue    = '-q '
    config.pbs.commands.submit  = 'qsub'
    config.pbs.commands.check  = 'qstat'
    config.pbs.commands.getid   = __pbs_get_jobid
    config.pbs.commands.running = __pbs_running
    # Slurm
    config.slurm.workdir = '-D'
    config.slurm.output  = '-o'
    config.slurm.error   = '-e'
    config.slurm.special.nodes    = '-N '
    config.slurm.special.cores    = '-n '
    config.slurm.special.walltime = '-t '
    config.slurm.special.memory = '--mem='
    config.slurm.special.queue    = '-p '
    config.slurm.commands.submit  = 'sbatch'
    config.slurm.commands.check  = 'squeue'
    config.slurm.commands.getid   = __slurm_get_jobid
    config.slurm.commands.running = __slurm_running


    def __init__(self, grid='auto', sleepstep=5, parallel=True, maxjobs=0):
        JobRunner.__init__(self, parallel=parallel, maxjobs=maxjobs)
        self.sleepstep = sleepstep
        self._active_jobs = {}
        self._active_lock = threading.Lock()
        self._mainlock = threading.Lock()

        if isinstance(grid, Settings):
            self.settings = grid
        elif grid == 'auto':
            self.settings = self._autodetect()
        elif grid in GridRunner.config:
            self.settings = GridRunner.config[grid]
            try:
                saferun([self.settings.commands.submit, '--version'], stdout=DEVNULL, stderr=DEVNULL)
            except OSError:
                raise PlamsError('GridRunner: {} command not found'.format(self.settings.commands.submit))
        else:
            raise PlamsError("GridRunner: invalid 'grid' argument. 'grid' should be either a Settings instance (see documentations for details) or a string occurring in GridRunner.config or 'auto' for autodetection")


    def call(self, runscript, workdir, out, err, runflags):
        """call(runscript, workdir, out, err, runflags)
        Submit *runscript* to the queueing system with *workdir* as the working directory. Redirect output and error streams to *out* and *err*, respectively. *runflags* stores varoius submit command options.

        The submit command has the following structure::

            <commands.submit>_<workdir>_{workdir}_<error>_{err}[_<output>_{out}][FLAGS]_{runscript}

        Underscores denote spaces, parts in pointy brackets correspond to ``settings`` entries, parts in curly brackets to :meth:`call` arguments, square brackets contain optional parts. Output part is added if *out* is not ``None``. This is handled automatically based on ``runscript.stdout_redirect`` value in job's ``settings``.

        ``FLAGS`` part is built based on *runflags* argument, which is a |Settings| instance storing |run| keyword arguments. For every *(key,value)* pair in *runflags* the string ``_-key_value`` is appended to ``FLAGS`` **unless** the *key* is a special key occurring in ``commands.special``. In that case ``_<commands.special.key>value`` is used (mind the lack of space in between). For example, a |Settings| instance defining interaction with SLURM has the following entries::

            workdir = '-D'
            output  = '-o'
            error   = '-e'
            special.nodes    = '-N '
            special.walltime = '-t '
            special.memory = '--mem='
            special.queue    = '-p '
            commands.submit  = 'sbatch'
            commands.check  = 'squeue'

        The submit command produced by::

            gr = GridRunner(parallel=True, maxjobs=4, grid='slurm')
            j.run(jobrunner=gr, queue='short', nodes=2, J='something', O='')

        will be:

        .. code-block:: none

            sbatch -D {workdir} -e {err} -o {out} -p short -N 2 -J something -O  {runscript}

        In certain queueing systems some flags don't have a short form with semantics ``-key value``. For example, in SLURM the flag ``--nodefile=value`` has a short form ``-F value``, but the flag ``--export=value`` does not. One can still use such a flag using the special keys logic::

            gr = GridRunner(parallel=True, maxjobs=4, grid='slurm')
            gr.settings.special.export = '--export='
            j.run(jobrunner=gr, queue='short', export='value')

        That results in the command::

            sbatch -D {workdir} -e {err} -o {out} -p short --export=value {runscript}

        The submit command is then executed and the output returned by it is used to determine the submitted job's ID. The value stored in ``commands.getid`` is used for that purpose. It should be a function taking a single string (the whole output of the submit command) and returning a string with job's ID.

        The submitted job's ID is then added to ``_active_jobs`` dictionary, with the key being job's ID and the value being an instance of :class:`threading.Lock`. This lock is used to singal the fact that the job is finished and the thread handling it can continue. Then the :meth:`_check_queue` method starts the thread querying the queue and unlocking finished jobs.

        Since it is difficult to automatically obtain job's exit code, the returned value is 0 (or 1, if the submit command failed). From |run| perspective it means that a job executed with |GridRunner| is *crashed* only if it never entered the queue (usually due to improper submit command).

        .. note::
            This method is used automatically during |run| and should never be explicitly called in your script.
        """
        s = self.settings
        cmd = ' '.join([s.commands.submit, s.workdir, workdir, s.error, err])
        if out is not None:
            cmd += ' '+s.output+' '+out
        for k,v in runflags.items():
            if k in s.special:
                cmd += ' '+s.special[k]+str(v)
            else:
                cmd += ' -'+k+' '+str(v)
        cmd += ' ' + opj(workdir,runscript)

        log('Submitting {} with command {}'.format(runscript, cmd), 5)
        process = saferun(cmd.split(' '), stdout=PIPE, stderr=PIPE)
        subout = process.stdout.decode()
        log('Output of {} submit command: {}'.format(runscript, subout), 5)

        jobid = s.commands.getid(subout)
        if jobid is None:
            log('Submitting of {} failed. Stderr of submit command:\n{}'.format(runscript, process.stderr.decode()), 1)
            return 1
        log('{} submitted successfully as job {}'.format(runscript, jobid), 3)

        event = threading.Event()
        with self._active_lock:
            self._active_jobs[jobid] = event
        self._check_queue()
        event.wait()

        log('Execution of {} finished'.format(runscript), 5)
        return 0


    @_in_thread
    def _check_queue(self):
        """Query the queueing system to obtain a list of currently running jobs. Check for active jobs that are not any more in the queue and release their locks. Repeat this procedure every ``sleepstep`` seconds until there are no more active jobs. The ``_mainlock`` lock ensures that there is at most one thread executing the main loop of this method at the same time."""
        if self._mainlock.acquire(blocking=False):
            try:
                while True:
                    with self._active_lock:
                        active_jobs = set(self._active_jobs.keys())

                    process = saferun([self.settings.commands.check], stdout=PIPE)
                    output = process.stdout.decode()
                    running_jobs = set(self.settings.commands.running(output))

                    with self._active_lock:
                        for jobid in active_jobs - running_jobs:
                            self._active_jobs[jobid].set()
                            del self._active_jobs[jobid]
                        if len(self._active_jobs) == 0:
                            return
                    time.sleep(self.sleepstep)
            finally:
                self._mainlock.release()


    def _autodetect(self):
        """Try to autodetect the type of queueing system.

        The autodetection mechanism is very simple. For each entry in ``GridRunner.config`` the submit command followed by ``--version`` is executed (for example ``qsub --version``). If the execution was successful (which is indicated by the exit code 0), that queueing system is present and it is chosen. Thus if there are multiple queueing systems installed, only one of them is picked -- the one which "name" (indicated by a key in ``GridRunner.config``) is first in the lexicographical order.

        Returned value is one of ``GridRunner.config`` branches. If autodetection was not successful, an exception is raised.
        """
        for grid in GridRunner.config:
            try:
                process = saferun([GridRunner.config[grid].commands.submit, '--version'], stdout=DEVNULL, stderr=DEVNULL)
            except OSError: continue
            if process.returncode == 0:
                log("Grid type autodetected as '{}'".format(grid), 5)
                return GridRunner.config[grid]
        raise PlamsError('GridRunner: Failed to autodetect grid type')



from ..interfaces.adfsuite.ams import AMSJob
from ..core.functions import log
import os

__all__ = ['ADFNBOJob']

class ADFNBOJob(AMSJob):

    def prerun(self):
        s = self.settings.input.ADF
        s.fullfock = True
        s.aomat2file = True
        s.symmetry = 'NoSym'
        s.basis.core = 'None'
        if 'save' in s:
            if isinstance(s.save, str):
                s.save += ' TAPE15'
            elif isinstance(s.save, list):
                s.save.append('TAPE15')
            else:
                log("WARNING: 'SAVE TAPE15' could not be added to the input settings of {}. Make sure (thisjob).settings.input.save is a string or a list.".format(self.name), 1)
        else:
            s.save = 'TAPE15'

        if isinstance(self.settings.adfnbo, list):
            adfnbo_input = self.settings.adfnbo
        else:
            adfnbo_input = ['write', 'spherical', 'fock']
            log('WARNING: (thisjob).settings.adfnbo should be a list. Using default settings: write, fock, spherical', 1)

        self.settings.runscript.post = 'cp "'+os.path.join(self.path,'adf.rkf')+'" TAPE21\n' '$AMSBIN/adfnbo <<eor\n' + '\n'.join(adfnbo_input) + '\neor\n\n$AMSBIN/gennbo6 FILE47\n'



from .scmjob import SCMJob, SCMResults

__all__ = ['UFFJob', 'UFFResults']



class UFFResults(SCMResults):
    _kfext = '.rkf'
    _rename_map = {'uff.rkf':'$JN'+_kfext}



class UFFJob(SCMJob):
    _result_type = UFFResults
    _command = 'uff'
    _top = ['title', 'units']
    _subblock_end = 'end'

    def _serialize_mol(self):
        s = self.settings.input

        printtypes = all(map(lambda at: ('uff' in at.properties and 'type' in at.properties.uff), self.molecule))

        for i,atom in enumerate(self.molecule):
            suffix = ''
            suffix_dict = {}
            if printtypes:
                suffix = '  {type}  {charge}'
                suffix_dict = atom.properties.uff.copy()
                if 'charge' not in suffix_dict:
                    suffix_dict.charge = 0.0
            s.system.atoms['_'+str(i+1)] = atom.str(suffix=suffix, suffix_dict=suffix_dict, space=18, decimal=10)

        if 'ignore_bonds' not in self.settings:
            self.molecule.set_atoms_id()
            for i,bond in enumerate(self.molecule.bonds):
                s.system.bonds['_'+str(i+1)] = '{:6d}  {:6d}  {:6.2f}'.format(bond.atom1.id, bond.atom2.id, bond.order)
            self.molecule.unset_atoms_id()

        if self.molecule.lattice:
            for i,vec in enumerate(self.molecule.lattice):
                s.system.lattice['_'+str(i+1)] = '{:16.10f} {:16.10f} {:16.10f}'.format(*vec)

    def _remove_mol(self):
        s = self.settings.input

        if 'system' in s:
            for k in ['atoms', 'bonds', 'lattice']:
                if k in s.system:
                    del s.system[k]



import numpy as np
from .units import Units
from ..interfaces.adfsuite.ams import AMSJob
from ..mol.molecule import Molecule
import os

__all__ = ['get_stoichiometry', 'balance_equation', 'reaction_energy']

def get_stoichiometry(job_or_molecule_or_path, as_dict=True):
    r = job_or_molecule_or_path
    d = None
    if isinstance(r, AMSJob):
        d = r.molecule.get_formula(as_dict=as_dict)
    elif isinstance(r, Molecule):
        d = r.get_formula(as_dict=as_dict)
    elif isinstance(r, dict):
        d = r.copy()
    elif isinstance(r, str):
        if os.path.isdir(r):
            d = AMSJob.load_external(r).molecule.get_formula(as_dict=as_dict)
        elif os.path.exists(r):
            try:
                d = Molecule(r).get_formula(as_dict=as_dict)
            except:
                d = AMSJob.load_external(r).molecule.get_formula(as_dict=as_dict)
        else:
            raise ValueError("The path {} does not exist.".format(r))
                
    else:
        raise TypeError("expected type AMSJob or dict but received {}".format(type(r)))

    return d

def balance_equation(reactants, products, normalization='r0'):
    """
    Calculate stoichiometric coefficients 
    This only works if 
        number_of_chemical_elements == len(reactants)+len(products), OR
        number_of_chemical_elements == len(reactants)+len(products)-1

    Returns: a 2-tuple (coeffs_reactants, coeffs_products)
        coeffs_reactants is a list with length == len(reactants)
        coeffs_products is a list with length == len(products)

    reactants: a list of amsjobs, or a list of paths to ams.results folders or ams.rkf files or .xyz files, or a list of Molecules, or a list of stoichiometry dicts, or a list of Molecules

    products: a list of amsjobs, or a list of paths to ams.results folders or ams.rkf files, or a list of Molecules or .xyz files, or a list of stoichiometry dicts, or a list of Molecules

    normalization: 'r0' for the first reactant, 'r1' for the second reactant, etc.
        'p0' for the first product, 'p1' for the second product, etc.
        This normalizes the chemical equation such that the coefficient in front of the specified species is 1

    EXAMPLE:

        balance_equation(
            reactants=[
                {'N': 2, 'H': 8, 'Cr': 2, 'O': 7}
            ],
            products=[
                {'Cr': 2, 'O': 3},
                {'N': 2},
                {'H': 2, 'O': 1}
            ])

        returns
        ([1.0], [1.0, 1.0, 1.0, 4.0])


    """
    def get_stoichiometries_and_elements(list_of_jobs):
        stoich = []
        elements = set()
        for r in list_of_jobs:
            d = get_stoichiometry(r)
            stoich.append(d)
            for k in d:
                elements.add(k)
        return stoich, elements

    def get_normalization_index(normalization):
        if normalization.startswith('r'):
            normalization_index = int(normalization.split('r')[1])
            if normalization_index >= num_reactants:
                raise ValueError("Reactant index {} specified, but max value allowed is {}".format(normalization_index, num_reactants-1))
        elif normalization.startswith('p'):
            normalization_index = int(normalization.split('p')[1]) 
            if normalization_index >= num_products:
                raise ValueError("Product index {} specified, but max value allowed is {}".format(normalization_index, num_products-1))
            normalization_index += num_reactants
        else:
            raise ValueError("Unknown normalization: {}. Should be r0, r1, r2, ... (for reactants), p0, p1, p2 ... (for products)")

        return normalization_index

    if len(reactants) == 0:
        raise ValueError('The reactants list is empty.')
    if len(products) == 0:
        raise ValueError('The products list is empty.')

    stoich_r, elements_r = get_stoichiometries_and_elements(reactants)
    stoich_p, elements_p = get_stoichiometries_and_elements(products)
    elements = elements_r
    elements.update(elements_p) #hopefully not necessary
    elements = list(elements)
    num_reactants = len(stoich_r)
    num_products = len(stoich_p)

    # EXAMPLE:
    # aCH4 + bO2 --> cCO2 + dH2O
    # mat =
    # [[-1 0    1 0], #C
    #  [-4 0    0 2], #H
    #  [0 -2    2 1]] #O
    #  CH4 O2 CO2 H2O
    #
    # Al2(SO4)3 + Ca(OH)2 → Al(OH)3 + CaSO4 
    # mat = np.array([[-2,-3,-12,0,0,0],[0,0,0,-1,-2,-2],[1,0,0,0,3,3],[0,1,4,1,0,0]]).T

    mat = []
    for e in elements:
        row = []
        for s in stoich_r:
            row.append(-s.get(e,0))
        for s in stoich_p:
            row.append(s.get(e,0))
        mat.append(row)
    mat = np.array(mat)

    coeffs = []
    if mat.shape[0] >= mat.shape[1]:
        u, s, vh = np.linalg.svd(mat)
        coeffs = np.compress(s <= 1e-12, vh, axis=0)
    elif mat.shape[0] == mat.shape[1]-1:
        # e.g. 3x4 matrix, so add a [1,1,1,1] row to find a particular solution
        newmat = np.concatenate((mat, np.array([[1]*mat.shape[1]])), axis=0)
        b = np.array([0]*(newmat.shape[0]-1)+[1.])
        try:
            coeffs = np.linalg.solve(newmat, b)
        except Exception as e:
            raise RuntimeError("Something went wrong when solving the system of linear equations. Verify that the chemical equation can be balanced at all, and that it can be balanced uniquely except for multiplication by a constant. {}\nA={}\nb={}".format(e, newmat,b))
    else:
        raise ValueError("The number of chemical elements must equal the number of molecules, or (the number of molecules-1). You have {} chemical elements: {}, and {} molecules".format(len(elements), elements, num_reactants+num_products))

    coeffs = coeffs.ravel()
    if len(coeffs) != num_reactants+num_products:
        raise RuntimeError("Something went wrong when solving the system of linear equations. Verify that the chemical equation can be balanced at all, and that it can be balanced uniquely except for multiplication by a constant.")

    normalization_index = get_normalization_index(normalization)
    coeffs /= coeffs[normalization_index]

    # double-check that the equation is balanced
    if abs(np.sum(mat @ coeffs.reshape(-1,1))) > 1e-10:
        raise RuntimeError('Stoichiometry double-check failed. mat = {}, coeffs = {}'.format(mat, coeffs))

    return list(coeffs[:num_reactants]), list(coeffs[num_reactants:])

def reaction_energy(reactants, products, normalization='r0', unit='hartree'):
    """ 

    Calculates a reaction energy from an unbalanced chemical equation (the equation is first balanced)

    reactants: a list of amsjobs or paths to ams results folders,
    products: a list of amsjobs or paths to ams results folders
    normalization: normalize the chemical equation by setting the corresponding coefficient to 1.
        'r0': first reactant
        'r1': second reactant, ...
        'p0: first product,
        'p1': second product, ...
    unit: Unit of the reaction energy

    Returns: a 3-tuple (coeffs_reactants, coeffs_products, reaction_energy)

    """

    my_reactants = [AMSJob.load_external(x) for x in reactants]
    my_products = [AMSJob.load_external(x) for x in products]
    coeffs_r, coeffs_p = balance_equation(my_reactants, my_products, normalization)
    reaction_energy = None

    energies = [job.results.get_energy(unit=unit) for job in my_reactants+my_products]
    energies = np.array(energies)

    coeffs = np.concatenate(([-x for x in coeffs_r],coeffs_p))
    reaction_energy = np.dot(coeffs, energies)

    return coeffs_r, coeffs_p, reaction_energy




from collections import OrderedDict
from ..core.functions import add_to_instance
from ..core.basejob import MultiJob
from ..core.results import Results
from ..core.settings import Settings
from ..mol.molecule import Molecule
from ..interfaces.adfsuite.ams import AMSJob

__all__ = ['ReorganizationEnergyJob', 'ReorganizationEnergyResults']


# using this function to pass a molecule inside a MultiJob ensures proper parallel execution
def pass_molecule(source, target):
    @add_to_instance(target)
    def prerun(self):
        self.molecule = source.results.get_main_molecule()


class ReorganizationEnergyResults(Results):
    """Results class for reorganization energy.
    """
    def reorganization_energy(self, unit='au'):
        energies = self.get_all_energies(unit)

        reorganization_energy = (energies['state B geo A'] - energies['state A geo A'] + 
                                 energies['state A geo B'] - energies['state B geo B'])
        return reorganization_energy

    def get_all_energies(self, unit='au'):
        energies = {}
        energies['state A geo A'] = self.job.children['go_A'].results.get_energy(unit=unit)
        energies['state B geo B'] = self.job.children['go_B'].results.get_energy(unit=unit)
        energies['state A geo B'] = self.job.children['sp_A_for_geo_B'].results.get_energy(unit=unit)
        energies['state B geo A'] = self.job.children['sp_B_for_geo_A'].results.get_energy(unit=unit)

        return energies


class ReorganizationEnergyJob(MultiJob):
    """A class for calculating the reorganization energy using AMS. 
    Given two states, A and B, the reorganization energy is defined as follows: 

    reorganization energy = 
    E(state B at optimal geometry for state A) - 
    E(state A at optimal geometry for state A) + 
    E(state A at optimal geometry for state B) - 
    E(state B at optimal geometry for state B)
    
    This job will run two geometry optimizations and two single point calculations.
    """

    _result_type = ReorganizationEnergyResults

    def __init__(self, molecule, common_settings, settings_state_A, settings_state_B, **kwargs):
        """
        molecule: the molecule 
        common_settings: a setting object for an AMSJob containing the shared settings for all the calculations
        settings_state_A: Setting object for an AMSJob containing exclusivelt the options defining the state A (e.g. charge and spin)
        settings_state_B: Setting object for an AMSJob containing exclusivelt the options defining the state B (e.g. charge and spin)
        kwargs: other options to be passed to the MultiJob constructor
        """
        MultiJob.__init__(self, children=OrderedDict(), **kwargs)

        go_settings = common_settings.copy()
        go_settings.input.ams.task = 'GeometryOptimization'

        sp_settings = common_settings.copy()
        sp_settings.input.ams.task = 'SinglePoint'

        # copy the settings so that we wont modify the ones provided as input by the user
        settings_state_A = settings_state_A.copy()
        settings_state_B = settings_state_B.copy()
        # In case the charge key is not specified, excplicitely set the value to 0.
        # This is to prevent the charge in molecule.properties.charge (set by get_main_molecule())
        # to be used in case of neutral systems
        for s in [settings_state_A, settings_state_B]:
            if not 'charge' in s.input.ams.system:
                s.input.ams.system.charge = 0

        self.children['go_A'] = AMSJob(molecule=molecule, settings=go_settings+settings_state_A, name='go_A')
        self.children['go_B'] = AMSJob(molecule=molecule, settings=go_settings+settings_state_B, name='go_B')
        self.children['sp_A_for_geo_B'] = AMSJob(settings=sp_settings+settings_state_A, name='sp_A_geo_B')
        self.children['sp_B_for_geo_A'] = AMSJob(settings=sp_settings+settings_state_B, name='sp_B_geo_A')
        pass_molecule(self.children['go_A'], self.children['sp_B_for_geo_A'])
        pass_molecule(self.children['go_B'], self.children['sp_A_for_geo_B'])




import numpy
import os
import shutil
import struct

from bisect import bisect
from collections import OrderedDict
from subprocess import DEVNULL

from ..core.private import saferun
from ..core.errors import FileError
from ..core.functions import log


__all__ = ['KFFile', 'KFReader', 'KFHistory']



class KFReader:
    """A class for efficient Python-native reader of binary files in KF format.

    This class offers read-only access to any fragment of data from a KF file. Unlike other Python KF readers, this one does not use the Fortran binary ``dmpkf`` to process KF files, but instead reads and interprets raw binary data straight from the file, on Python level. That approach results in significant speedup (by a factor of few hundreds for large files extracted variable by variable).

    The constructor argument *path* should be a string with a path (relative or absolute) to an existing KF file.

    *blocksize* indicates the length of basic KF file block. So far, all KF files produced by any of Amsterdam Modeling Suite programs have the same block size of 4096 bytes. Unless you're doing something *very* special, you should not touch this value.

    Organization of data inside KF file can depend on a machine on which this file was produced. Two parameters can vary: the length of integer (32 or 64 bit) and endian (little or big). These parameters have to be determined before any reading can take place, otherwise the results will have no sense. If the constructor argument *autodetect* is ``True``, the constructor attempts to automatically detect the format of a given KF file, allowing to read files created on a machine with different endian or integer length. This automatic detection is enabled by default and it is advised to leave it that way. If you wish to disable it, you should set ``endian`` and ``word`` attributes manually before reading anything (see the code for details).

    .. note ::

        This class consists of quite technical, low level code. If you don't need to modify or extend |KFReader|, you can safely ignore all private methods, all you need is :meth:`~KFReader.read` and occasionally :meth:`~KFReader.__iter__`

    """

    _sizes = {'s':1,'i':4,'d':8,'q':8}

    def __init__(self, path, blocksize=4096, autodetect=True):
        if os.path.isfile(path):
            self.path = os.path.abspath(path)
        else:
            raise FileError('File {} not present'.format(path))

        self._blocksize = blocksize
        self.endian = '<'   # endian: '<' = little, '>' = big
        self.word = 'i'     # length of int: 'i' = 4 bits, 'q' = 8 bits
        self._sections = None
        if autodetect:
            self._autodetect()


    def read(self, section, variable):
        """Extract and return data for a *variable* located in a *section*.

        For single-value numerical or boolean variables returned value is a single number or bool. For longer variables this method returns a list of values. For string variables a single string is returned.
        """

        if self._sections is None:
            self._create_index()

        try:
            tmp = self._sections[section]
        except KeyError:
            raise KeyError('Section {} not present in {}'.format(section, self.path))
        try:
            vtype, vlb, vstart, vlen = tmp[variable]
        except KeyError:
            raise KeyError('Variable {} not present in section {} of {}'.format(variable, section, self.path))

        ret = []
        first = True
        with open(self.path, 'rb') as f:
            for i in KFReader._datablocks(self._data[section], vlb):
                if first:
                    ret = self._get_data(self._read_block(f,i), vtype)[vstart-1:]
                    first = False
                else:
                    ret += self._get_data(self._read_block(f,i), vtype)
                if len(ret) >= vlen:
                    ret = ret[:vlen]
                    if isinstance(ret, bytes):
                        try:
                            return ret.decode()
                        except UnicodeDecodeError:
                            return ret.decode("Latin-1")
                    elif len(ret) == 1:
                        return ret[0]
                    else:
                        return ret


    def __iter__(self):
        """Iteration yields pairs of section name and variable name."""
        if self._sections is None:
            self._create_index()
        for section in self._sections:
            for variable in self._sections[section]:
                yield section, variable


    def _autodetect(self):
        """Try to automatically detect the format (int size and endian) of this KF file."""
        with open(self.path, 'rb') as f:
            b = f.read(128)

        blocksize = struct.unpack(b'i',b[28:32])[0]
        self._blocksize = 4096 if blocksize == 538976288 else blocksize
        log('Block size of {} detected as {}'.format(self.path, self._blocksize), 7)

        one = b[80:84]

        if struct.unpack(b'32s',b[48:80])[0] == b'SUPERINDEX                      ':
            self.word = 'i'
        elif struct.unpack(b'32s',b[64:96])[0] == b'SUPERINDEX                      ':
            self.word = 'q'
            one = b[96:104]
        else:
            log('WARNING: Unable to autodetect integer size and endian of {}. Using defaults (4 bytes and little endian)'.format(self.path), 3)
            return

        for e in ['<', '>']:
            if struct.unpack(str(e+self.word), one)[0] == 1:
                self.endian = e
                d = {'q':'8 bytes','i':'4 bytes','<':'little endian','>':'big endian',}
                log(('Format of {0} detected to {'+self.word+'} and {'+self.endian+'}').format(self.path, **d), 7)


    def _read_block(self, f, pos):
        """Read a single block of binary data from posistion *pos* in file *f*."""
        f.seek((pos-1)*self._blocksize)
        return f.read(self._blocksize)


    def _parse(self, block, format):   #format = [(32,'s'),(4,'i'),(2,'d')]
        """Translate a *block* of binary data into list of values in specified *format*.

        *format* should be a list of pairs *(a,t)* where *t* is one of the following characters: ``'s'`` for string (bytes), ``'i'`` for 32-bit integer, ``'q'`` for 64-bit integer and *a* is the number of occurrences (or length of a string).

        For example, if *format* is equal to ``[(32,'s'),(4,'i'),(2,'d'),(2,'i')]``, the contents of *block* are divided into 72 bytes (32*1 + 4*4 + 2*8 + 2*4 = 72) chunks (possibly droping the last one, if it's shorter than 72 bytes). Then each chunk is translated to a 9-tuple of bytes, 4 ints, 2 floats and 2 ints. List of such tuples is the returned value.
        """
        step = 0
        formatstring = self.endian
        for a,t in format:
            step += a * self._sizes[t]
            formatstring += str(a) + t

        if step > 0:
            end = (len(block) // step) * step
            return list(struct.iter_unpack(formatstring, block[:end]))
        else:
            return []


    def _get_data(self, datablock, vtype):
        """Extract all data of a given type from a single data block. Returned value is a list of values (int, float, or bool) or a single "bytes" object.
        """
        hlen = 4 * self._sizes[self.word]
        i,d,s,b = self._parse(datablock[:hlen],[(4,self.word)])[0]
        contents = self._parse(datablock[hlen:], zip((i,d,s,b),(self.word,'d','s',self.word)))
        if contents:
            contents = contents[0]  # there won't be more than one chunk of data in any data block
            if vtype == 1:
                return list(contents[:i])
            elif vtype == 2:
                return list(contents[i:i+d])
            elif vtype == 3:
                return contents[i+d]
            elif vtype == 4:
                return list(map(bool,contents[i+d+1:]))
            else:
                raise KeyError('Unknown vtype')
        else:
            return []


    def _create_index(self):
        """Find and parse relevant index blocks of KFFile to extract the information about location of all sections and variables.

        Two dictionaries are populated during this process. ``_data`` contains, for each section, a list of triples describing how logical blocks of data are mapped into physical ones. For example, ``_data['General'] = [(3,6,12), (9,40,45)]`` means that logical blocks 3-8 of section ``General`` are located in physical blocks 6-11 and logical blocks 9-13 in physical blocks 40-44. This list is always sorted via first tuple elements allowing efficient access to arbitrary logical block of each section.

        The second dictionary, ``_sections``, is used to locate each variable within its section. For each section, it contains another dictionary of each variable of this section. So ``_section[sec][var]`` contains all information needed to extract variable ``var`` from section ``sec``. This is a 4-tuple containing the following information: variable type, logic block in which the variable first occurs, position within this block where its data start and the length of the variable. Combining this information with mapping stored in ``_data`` allows to extract each single variable.
        """

        hlen = 32 + 7 * self._sizes[self.word]   #length of index block header

        with open(self.path, 'rb') as f:
            superlist = self._parse(self._read_block(f, 1), [(32,'s'),(4,self.word)])
            nextsuper = superlist[0][4]
            while nextsuper != 1:
                nsl = self._parse(self._read_block(f, nextsuper), [(32,'s'),(4,self.word)])
                nextsuper = nsl[0][4]
                superlist += nsl

            self._data = {}   #list of triples to convert logical to physical block numbers
            self._sections = {}
            for key, pb, lb, le, ty in superlist:   #pb=physical block, lb=logical block, le=length, ty=type (3 for index, 4 for data)
                try:
                    key = key.decode()
                except UnicodeDecodeError:
                    key = key.decode("Latin-1")
                key = key.rstrip(' ')
                if key in ['SUPERINDEX', 'EMPTY']: continue
                if ty == 4:   #data block
                    if key not in self._data:
                        self._data[key] = []
                    self._data[key].append((lb, pb, pb+le))
                elif ty == 3:   #index block
                    if key not in self._sections:
                        self._sections[key] = {}
                    for i in range(le):
                        indexblock = self._read_block(f, pb+i)
                        body = self._parse(indexblock[hlen:],[(32,'s'),(6,self.word)])
                        for var, vlb, vstart, vlen, _xx1, vused, vtype in body:
                            try:
                                var = var.decode()
                            except UnicodeDecodeError:
                                var = var.decode("Latin-1")
                            var = var.rstrip(' ')
                            if var == 'EMPTY': continue
                            self._sections[key][var] = (vtype, vlb, vstart, vused)

            for k,v in self._data.items():
                lbs = []
                pbs = []
                for lb, first, last in sorted(v):
                    lbs.append(lb)
                    pbs.append((first, last))
                self._data[k] = (lbs, pbs)


    @staticmethod
    def _datablocks(lst, n=1):
        """Transform a tuple of lists ``([x1,x2,...], [(a1,b1),(a2,b2),...])`` into an iterator over ``range(a1,b1)+range(a2,b2)+...`` Iteration starts from nth element of this list."""
        i = bisect(lst[0], n) - 1
        lb = lst[0][i]
        first, last = lst[1][i]
        ret = first + n - lb
        while i < len(lst[1]):
            while ret < last:
                yield ret
                ret += 1
            i += 1
            if i < len(lst[1]):
                ret, last = lst[1][i]



#===========================================================================
#===========================================================================
#===========================================================================



class KFFile:
    """A class for reading and writing binary files in KF format.

    This class acts as a wrapper around |KFReader| collecting all the data written by user in some "temporary zone" and using Fortran binaries ``udmpkf`` and ``cpkf`` to write this data to the physical file when needed.

    The constructor argument *path* should be a string with a path to an existing KF file or a new KF file that you wish to create. If a path to existing file is passed, new |KFReader| instance is created allowing to read all the data from this file.

    When :meth:`~KFFile.write` method is used, the new data is not immediately written to a disk. Instead of that, it is temporarily stored in ``tmpdata`` dictionary. When method :meth:`~KFFile.save` is invoked, contents of that dictionary are written to a physical file and ``tmpdata`` is emptied.

    Other methods like :meth:`~KFFile.read` or :meth:`~KFFile.delete_section` are aware of ``tmpdata`` and work flawlessly, regardless if :meth:`~KFFile.save` was called or not.

    By default, :meth:`~KFFile.save` is automatically invoked after each :meth:`~KFFile.write`, so physical file on a disk is always "actual". This behavior can be adjusted with *autosave* constructor parameter. Having autosave enabled is usually a good idea, however, if you need to write a lot of small pieces of data to your file, the overhead of calling ``udmpkf`` and ``cpkf`` after *every* :meth:`~KFFile.write` can lead to significant delays. In such a case it is advised to disable autosave and call :meth:`~KFFile.save` manually, when needed.

    Dictionary-like bracket notation can be used as a shortcut to read and write variables::

        mykf = KFFile('someexistingkffile.kf')
        #all three below are equivalent
        x = mykf['General%Termination Status']
        x = mykf[('General','Termination Status')]
        x = mykf.read('General','Termination Status')

        #all three below are equivalent
        mykf['Geometry%xyz'] = somevariable
        mykf[('Geometry','xyz')] = somevariable
        mykf.write('Geometry','xyz', somevariable)

    """
    _types = {int : (1, 8, lambda x:'%10i'%x),
            float : (2, 3, lambda x:'%26.16e'%x),
              str : (3, 80, lambda x: x),
             bool : (4, 80, lambda x: 'T' if x else 'F')}


    def __init__(self, path, autosave=True):
        self.autosave = autosave
        self.path = os.path.abspath(path)
        self.tmpdata = OrderedDict()
        self.reader = KFReader(self.path) if os.path.isfile(self.path) else None


    def read(self, section, variable, return_as_list=False):
        """Extract and return data for a *variable* located in a *section*.

        By default, for single-value numerical or boolean variables returned value is a single number or bool. For longer variables this method returns a list of values. For string variables a single string is returned. This behavior can be changed by setting *return_as_list* parameter to ``True``. In that case the returned value is always a list of numbers (possibly of length 1) or a single string.
        """
        if section in self.tmpdata and variable in self.tmpdata[section]:
            ret = self.tmpdata[section][variable]
        else:
            ret = self.reader.read(section, variable)
        if return_as_list and isinstance(ret, (int,float,bool)):
            ret = [ret]
        return ret


    def write(self, section, variable, value):
        """Write a *variable* with a *value* in a *section* . If such a variable already exists in this section, the old value is overwritten."""
        if not isinstance(value, (int,bool,float,str,list)):
            raise ValueError('Trying to store improper value in KFFile')
        if isinstance(value, list):
            if len(value) == 0:
                raise ValueError('Cannot store empty lists in KFFile')
            if any(not isinstance(i, type(value[0])) for i in value):
                raise ValueError('Lists stored in KFFile must have all elements of the same type')
            if not isinstance(value[0], (int,bool,float,str)):
                raise ValueError('Only lists of int, float, str or bool can be stored in KFFile')

        if section not in self.tmpdata:
            self.tmpdata[section] = OrderedDict()
        self.tmpdata[section][variable] = value

        if self.autosave:
            self.save()


    def save(self):
        """Save all changes stored in ``tmpdata`` to physical file on a disk."""
        if len(self.tmpdata) > 0 and any(len(i) > 0 for i in self.tmpdata.values()):
            txt = ''
            newvars = []
            for section in self.tmpdata:
                for variable in self.tmpdata[section]:
                    val = self.tmpdata[section][variable]
                    txt += '{}\n{}\n{}\n'.format(section, variable, KFFile._str(val))
                    newvars.append(section+'%'+variable)
            self.tmpdata = OrderedDict()

            tmpfile = self.path+'.tmp' if self.reader else self.path
            saferun(['udmpkf', tmpfile], input=txt.encode('iso8859'), stdout=DEVNULL, stderr=DEVNULL)
            if self.reader:
                saferun(['cpkf', tmpfile, self.path] + newvars, stdout=DEVNULL, stderr=DEVNULL)
                os.remove(tmpfile)
            self.reader = KFReader(self.path)


    def delete_section(self, section):
        """Delete the entire *section* from this KF file."""
        if section in self.tmpdata:
            del self.tmpdata[section]
        if self.reader:
            if not self.reader._sections:
                self.reader._create_index()
            if section in self.reader._sections:
                tmpfile = self.path+'.tmp'
                saferun(['cpkf', self.path, tmpfile, '-rm', section], stdout=DEVNULL, stderr=DEVNULL)
                shutil.move(tmpfile, self.path)
                self.reader = KFReader(self.path)


    def sections(self):
        """Return a list with all section names, ordered alphabetically."""
        ret = set(self.tmpdata)
        if self.reader:
            if self.reader._sections is None:
                self.reader._create_index()
            ret |= set(self.reader._sections)
        ret = list(ret)
        ret.sort()
        return ret


    def read_section(self, section):
        """Return a dictionary with all variables from a given *section*.

        .. note::

            Some sections can contain very large amount of data. Turning them into dictionaries can cause memory shortage or performance issues. Use this method carefully.

        """
        ret = {}
        for sec, var in self:
            if sec == section:
                ret[var] = self.read(sec, var)
        if len(ret) == 0:
            log("WARNING: Section '{}' not present in {} or present, but empty. Returning empty dictionary".format(section, self.path), 1)
        return ret


    def get_skeleton(self):
        """Return a dictionary reflecting the structure of this KF file. Each key in that dictionary corresponds to a section name of the KF file with the value being a set of variable names."""
        ret = {}
        for sec,var in self:
            if sec not in ret:
                ret[sec] = set()
            ret[sec].add(var)
        return ret


    def __getitem__(self, name):
        """Allow to use ``x = mykf['section%variable']`` or ``x = mykf[('section','variable')]`` instead of ``x = kf.read('section', 'variable')``."""
        section, variable = KFFile._split(name)
        return self.read(section, variable)


    def __setitem__(self, name, value):
        """Allow to use ``mykf['section%variable'] = value`` or ``mykf[('section','variable')] = value`` instead of ``kf.write('section', 'variable', value)``."""
        section, variable = KFFile._split(name)
        self.write(section, variable, value)


    def __iter__(self):
        """Iteration yields pairs of section name and variable name."""
        ret = set()
        if self.reader:
            for sec,var in self.reader:
                ret.add((sec,var))
        for sec in self.tmpdata:
            for var in self.tmpdata[sec]:
                ret.add((sec,var))
        ret = list(ret)
        ret.sort(key=lambda x: x[0]+x[1])
        for i in ret:
            yield i


    def __contains__(self, arg):
        """Implements Python ``in`` operator for KFFiles. *arg* can be a single string with a section name or a pair of strings (section, variable)."""
        if isinstance(arg, str):
            return arg in self.sections()
        if isinstance(arg, tuple) and len(arg) == 2 and isinstance(arg[0],str) and isinstance(arg[1],str):
            try:
                self.read(*arg)
                return True
            except KeyError:
                return False
        raise TypeError("'in <KFFile>' requires string of a pair of strings as left operand")


    @staticmethod
    def _split(name):
        """Ensure that a key used in bracket notation is of the form ``'section%variable'`` or ``('section','variable')``. If so, return a tuple ``('section','variable')``."""
        if isinstance(name, tuple) and len(name) == 2:
                return name[0], name[1]
        if isinstance(name, str):
            s = name.split('%')
            if len(s) == 2:
                return s[0], s[1]
        raise ValueError('Improper key used in KFFile dictionary-like notation')


    @staticmethod
    def _str(val):
        """Return a string representation of *val* in the form that can be understood by ``udmpkf``."""
        if isinstance(val, (int,bool,float)):
            val = [val]
        valtype = type(val[0])
        t,step,f = KFFile._types[valtype]
        l = len(val)
        if (valtype == str and isinstance(val, list)):
            #udmpkf reads 160 characters per variable, split over max. 80 per line, to make a string array
            l = l * 160
            step = 1
            splitstrings = [[s[0:80], s[80:160]] for s in val]
            val = [item for sublist in splitstrings for item in sublist]

        ret = '%10i%10i%10i'%(l,l,t)
        for i,el in enumerate(val):
            if i%step == 0: ret += '\n'
            ret += f(el)
        return ret


#===========================================================================
#===========================================================================
#===========================================================================



class KFHistory:
    """A class for reading "History" sections of files in the KF format.

    This class acts as a wrapper around |KFReader| enabling convenient iteration over entries (frames) of History sections.

    The constructor argument *kf* should be a |KFReader| instance attached to an existing KF file. The *section* argument then holds a name of the desired History-like section, such as "History" or "MDHistory".

    The :meth:`~KFHistory.read_all` method can be used used to easily read all values of a particular history item into a single numpy array.

    To iterate over the frames in a history section, use :meth:`~KFHistory.iter` or :meth:`~KFHistory.iter_optional`. The former raises an exception if the selected variable is not present in the history, while the latter returns a given default value instead.

    Usage::

        kf = KFReader('somefile.rkf')
        mdhistory = KFHistory(kf, 'MDHistory')

        for T, p in mdhistory.iter('Temperature'), mdhistory.iter_optional('Pressure', 0):
            print(T, p)

    """

    def __init__(self, kf, section):
        self.kf = kf
        self.section = section
        self.nsteps = kf.read(section, "nEntries")
        self.shapes = {}
        self.blocked = set()

        if (section, "nBlocks") in kf:
            self.nblocks = kf.read(section, "nBlocks")
        else:
            self.nblocks = 0

    def read_all(self, name):
        """Return a numpy array containing the values of history item *name* from all frames."""
        if name not in self.shapes:
            self._init_shape(name)
        if name in self.blocked:
            return numpy.concatenate([numpy.atleast_1d(self.kf.read(self.section, "{}({})".format(name, i))) for i in range(1, self.nblocks + 1)])
        else:
            return numpy.asarray([self.kf.read(self.section, "{}({})".format(name, i)) for i in range(1, self.nsteps + 1)])

    def iter(self, name):
        """Iterate over the values of history item *name*."""
        if name not in self.shapes:
            self._init_shape(name)
        if name in self.blocked:
            for i in range(1, self.nblocks + 1):
                block = self.kf.read(self.section, "{}({})".format(name, i))
                try:
                    yield from block
                except TypeError:
                    # one-element blocks are not iterable (KFReader returns them as scalars)
                    yield block
        else:
            for i in range(1, self.nsteps + 1):
                yield self.kf.read(self.section, "{}({})".format(name, i))

    def iter_optional(self, name, default=None):
        """Iterate over the values of history item *name*, returning *default* if the item is not present."""
        if (self.section, name + "(1)") in self.kf:
            yield from self.iter(name)
        else:
            while True:
                yield default

    def _init_shape(self, name):
        shapevar = name + "(dim)"
        if (self.section, shapevar) in self.kf:
            shape = self.kf.read(self.section, shapevar)
            try:
                # shape is a list (variable "name" is at least rank-2)
                self.shapes[name] = tuple(shape)
            except TypeError:
                # shape is a scalar (variable "name" is a scalar or rank-1)
                self.shapes[name] = (shape,)
            perAtomVar = name + "(perAtom)"
            if self.nblocks and (self.section, perAtomVar) in self.kf:
                perAtom = self.kf.read(self.section, perAtomVar)
                if perAtom == False and self.shapes[name] == (1,):
                    self.blocked.add(name)
        else:
            # no shape info => assume scalar/rank-1 (read as is, no blocks)
            self.shapes[name] = (1,)



import multiprocessing

# Run as many job simultaneously as there are cpu on the system:
maxjobs = multiprocessing.cpu_count()
print('Running up to {} jobs in parallel simultaneously'.format(maxjobs))

# Set the default jobrunner to be parallel:
config.default_jobrunner = JobRunner(parallel=True, maxjobs=maxjobs)

# Number of cores for each job:
config.job.runscript.nproc = 1

# Load all molecules in the folder 'molecules':
molecules = read_molecules('molecules')

settings = Settings()
settings.input.ams.Task = 'GeometryOptimization'
settings.input.dftb.Model = 'GFN1-xTB'

results = []
for name, molecule in sorted(molecules.items()):
    job = AMSJob(molecule=molecule, settings=settings, name=name)
    results.append(job.run())

# Only print the results of the succesful caluclations:
for result in [r for r in results if r.ok()]:
    print('Energy for {:<12}: {:>10.3f} kcal/mol'.format(result.name, result.get_energy(unit='kcal/mol')))



def __autoimport(path, folders):
    import os
    from os.path import join as opj
    is_module = lambda x: x.endswith('.py') and not x.startswith('__init__')

    ret = []
    for folder in folders:
        for dirpath, dirnames, filenames in os.walk(opj(path,folder)):
            modules = [os.path.splitext(f)[0] for f in filter(is_module, filenames)]
            relpath = os.path.relpath(dirpath, path).split(os.sep)
            for module in modules:
                imp = '.'.join(relpath + [module])
                tmp = __import__(imp, globals=globals(), fromlist=['*'], level=1)
                if hasattr(tmp, '__all__'):
                    ret += tmp.__all__
                    for name in tmp.__all__:
                        globals()[name] = vars(tmp)[name]
    return ret


__all__ = __autoimport(__path__[0], ['core', 'mol', 'interfaces', 'tools', 'recipes','trajectories'])

__version__ = '1.5.1'
